# -*- coding: utf-8 -*-

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = ['sphinx.ext.autodoc', 'sphinx.ext.todo']

# Add any paths that contain templates here, relative to this directory.
templates_path = ['templates']

# The suffix of source filenames.
source_suffix = '.txt'

# The encoding of source files.
#source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'pymta'
copyright = '2008-2011 Felix Schwarz'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = '0.5'
# The full version, including alpha/beta/rc tags.
release = '0.5dev'

# List of directories, relative to source directory, that shouldn't be searched
# for source files.
exclude_trees = []

# The reST default role (used for this markup: `text`) to use for all documents.
#default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
add_function_parentheses = True

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'


# -- Options for HTML output ---------------------------------------------------

# The theme to use for HTML and HTML Help pages.  Major themes that come with
# Sphinx are currently 'default' and 'sphinxdoc'.
html_theme = 'default'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = []


# Output file base name for HTML help builder.
htmlhelp_basename = 'pymtadoc'


# -- Options for LaTeX output --------------------------------------------------

# The paper size ('letter' or 'a4').
#latex_paper_size = 'letter'

# The font size ('10pt', '11pt' or '12pt').
#latex_font_size = '10pt'

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]).
latex_documents = [
  ('index', 'pymta.tex', u'pymta Documentation',
   u'Felix Schwarz', 'manual'),
]



# -*- coding: UTF-8 -*-
# SPDX-License-Identifier: MIT


from pymta.api import PyMTAException


__all__ = ['InvalidParametersError', 'SMTPViolationError']


class SMTPViolationError(PyMTAException):
    """Raised when the SMTP client violated the protocol."""

    def __init__(self, response_sent=False, code=550,
                 reply_text='Administrative Prohibition', message=None):
        if message is None:
            message = '%s %s' % (code, reply_text)
        PyMTAException.__init__(self, message)
        self.response_sent = response_sent
        self.code = code
        self.reply_text = reply_text


class InvalidParametersError(SMTPViolationError):
    """The SMTP client provided invalid parameters for a SMTP command."""

    def __init__(self, parameter=None, *args, **kwargs):
        # In Python 2.3 Exception is an old-style classes so we can not use super
        SMTPViolationError.__init__(self, *args, **kwargs)
        self.parameter = parameter





# -*- coding: UTF-8 -*-
# SPDX-License-Identifier: MIT

from threading import Event
import socket
import time

from pymta.command_parser import WorkerProcess

__all__ = ['PythonMTA']



def forked_child(queue, server_socket, deliverer_class, policy_class,
                 authenticator_class):
    child = WorkerProcess(queue, server_socket, deliverer_class, policy_class,
                          authenticator_class)
    child.run()



class PythonMTA(object):
    """Create a new MTA which listens for new connections afterwards.
    local_address is a string containing either the IP oder the DNS
    host name of the interface on which PythonMTA should listen.
    deliverer_class, policy_class and authenticator_class are callables which
    can be used to add custom behavior. Please note that they must be picklable
    if you use forked worker processes (default).
    Every new connection gets their own instance of policy_class and
    authenticator_class so these classes don't have to be thread-safe. If
    you omit the policy, all syntactically valid SMTP commands are
    accepted. If there is no authenticator specified, authentication will
    not be available."""

    def __init__(self, local_address, bind_port, deliverer_class,
                 policy_class=None, authenticator_class=None):
        self._local_address = local_address
        self._bind_port = bind_port
        self._deliverer_class = deliverer_class
        self._policy_class = policy_class
        self._authenticator_class = authenticator_class

        self._queue = None
        self._processes = []
        self._shutdown_server = Event()

    def _try_to_bind_to_socket(self, server_socket):
        tries = 0
        while tries < 10:
            try:
                server_socket.bind((self._local_address, self._bind_port))
            except socket.error:
                tries += 1
                time.sleep(0.1)
            else:
                break

    def _build_server_socket(self):
        server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        # If the server crashed and we restarted it within a very short time
        # frame, prevent 'address already in use' errors.
        server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        # We want to terminate all children within a reasonable time
        server_socket.settimeout(1)
        self._try_to_bind_to_socket(server_socket)
        # Don't loose connections in the time frame when a new connection was
        # accepted. Python's documentation says the maximum is system dependent
        # but usually 5 so we take that.
        server_socket.listen(5)
        return server_socket

    def _get_child_args(self, server_socket):
        return (self._queue, server_socket, self._deliverer_class,
                self._policy_class, self._authenticator_class)

    def _start_new_worker_process(self, server_socket):
        """Start a new child worker process which will listen on the given
        socket and return a reference to the new process."""
        from multiprocessing import Process
        p = Process(target=forked_child, args=self._get_child_args(server_socket))
        p.start()
        return p

    def serve_forever(self, use_multiprocessing=True):
        if use_multiprocessing:
            try:
                from multiprocessing import Queue
            except ImportError:
                use_multiprocessing = False
        if not use_multiprocessing:
            from pymta.compat import queue
            Queue = queue.Queue

        self._shutdown_server.clear()
        self._queue = Queue()
        # Put the initial token in the Queue
        self._queue.put(True)
        server_socket = self._build_server_socket()
        if use_multiprocessing:
            for i in range(5):
                p = self._start_new_worker_process(server_socket)
                self._processes.append(p)
            while not self._shutdown_server.isSet():
                time.sleep(1)
            for process in self._processes:
                process.join()
        else:
            forked_child(*self._get_child_args(server_socket))
        server_socket.close()
        self._queue = None

    def shutdown_server(self, timeout_seconds=None):
        """This method notifies the server that it should stop listening for
        new messages and shut down itself. If timeout_seconds was given, the
        method will block for this many seconds at most."""
        self._queue.put(None)
        self._shutdown_server.set()
        # TODO: Looks like we're quitting too fast here.





# -*- coding: UTF-8 -*-
# SPDX-License-Identifier: MIT

from pycerberus.errors import InvalidDataError
from pythonic_testcase import *

from pymta.validation import SMTPEmailValidator


class SMTPEmailValidatorTest(PythonicTestCase):

    def process(self, input_string):
        return SMTPEmailValidator().process(input_string)

    def test_accept_plain_email_address(self):
        self.assert_equals('foo@example.com', self.process('foo@example.com'))

    def test_accept_email_address_in_angle_brackets(self):
        self.assert_equals('foo@example.com', self.process('<foo@example.com>'))

    def test_reject_email_address_in_unbalanced_angle_brackets(self):
        self.assert_raises(InvalidDataError, lambda: self.process('<foo@example.com'))
        self.assert_raises(InvalidDataError, lambda: self.process('foo@example.com>'))

    def test_reject_invalid_email_addresses(self):
        self.assert_raises(InvalidDataError, lambda: self.process('foo@@example.com'))
        self.assert_raises(InvalidDataError, lambda: self.process('foo@example..com'))





#!/usr/bin/env python
# -*- coding: UTF-8 -*-
# SPDX-License-Identifier: MIT
"""This module contains an equivalent of Python's DebuggingServer which just
prints all received messages to STDOUT and discards them afterwards."""

import os
import sys

from pymta import PythonMTA, IMessageDeliverer


class STDOUTDeliverer(IMessageDeliverer):
    def new_message_accepted(self, msg):
        print '---------- MESSAGE FOLLOWS ----------'
        print msg.msg_data
        print '------------ END MESSAGE ------------'


def list_get(data, index, default=None):
    if len(data) <= index:
        return default
    return data[index]


def print_usage():
    cmd_name = list_get(sys.argv, 0, default=os.path.basename(__file__))
    print 'Usage: %s [host] [port]' % cmd_name


if __name__ == '__main__':
    if len(sys.argv) < 2:
        print_usage()
        sys.exit(0)

    host = list_get(sys.argv, 1, default='localhost')
    port = int(list_get(sys.argv, 2, default=8025))

    server = PythonMTA(host, port, STDOUTDeliverer)
    try:
        server.serve_forever()
    except KeyboardInterrupt:
        pass




# -*- coding: UTF-8 -*-
# SPDX-License-Identifier: MIT

from pymta.api import IMTAPolicy, PolicyDecision

from tests.util import CommandParserTestCase


class PolicyReturnCodesTest(CommandParserTestCase):
    "Tests the different outputs of a policy class"

    def test_policy_can_return_true(self):
        class TruePolicy(IMTAPolicy):
            def accept_new_connection(self, peer):
                return True
        self.init(policy=TruePolicy())
        self.assert_true(self.command_parser.open)

    def test_policy_can_return_false(self):
        class FalsePolicy(IMTAPolicy):
            def accept_new_connection(self, peer):
                return False
        self.init(policy=FalsePolicy())
        self.assert_false(self.command_parser.open)

    def test_returning_none_is_treated_as_true(self):
        class NonePolicy(IMTAPolicy):
            def accept_new_connection(self, peer):
                return None
        self.init(policy=NonePolicy())
        self.assert_true(self.command_parser.open)

    def test_policy_can_return_custom_codes_as_tuple(self):
        class CustomCodePolicy(IMTAPolicy):
            def accept_new_connection(self, peer):
                return (False, (553, 'Go away'))
        self.init(policy=CustomCodePolicy())

        self.assert_false(self.command_parser.open)
        self.assert_equals(1, len(self.command_parser.replies))
        code, reply_text = self.command_parser.replies[-1]
        self.assert_equals(553, code)
        self.assert_equals('Go away', reply_text)

    def test_policy_can_return_multiple_lines(self):
        class CustomCodePolicy(IMTAPolicy):
            def accept_new_connection(self, peer):
                return (False, (552, ('Go away', 'Evil IP')))
        self.init(policy=CustomCodePolicy())

        self.assert_equals(1, len(self.command_parser.replies))
        code, reply_text = self.command_parser.replies[-1]
        self.assert_equals(552, code)
        self.assert_equals(('Go away', 'Evil IP'), reply_text)
        self.assert_false(self.command_parser.open)

    def test_can_return_policydecision_instance(self):
        class ReturnPolicyDecisionPolicy(IMTAPolicy):
            def accept_helo(self, helo_string, message):
                is_localhost = (helo_string == 'localhost')
                return PolicyDecision(decision=is_localhost)
        self.init(policy=ReturnPolicyDecisionPolicy())

        self.send('HELO', 'foo.example.com', expected_first_digit=5)
        self.send('HELO', 'bar.example.net', expected_first_digit=5)
        self.send('HELO', 'localhost')

    def test_can_return_custom_code_and_message_in_policydecision_instance(self):
        class ReturnCustomCodePolicy(IMTAPolicy):
            def accept_helo(self, helo_string, message):
                return PolicyDecision(False, (553, 'I am tired'))
        self.init(policy=ReturnCustomCodePolicy())

        self.send('HELO', 'foo.example.com', expected_first_digit=5)
        code, reply_text = self.command_parser.replies[-1]
        self.assert_equals((553, 'I am tired'), (code, reply_text))

    def test_can_close_connection_after_reply(self):
        class CloseConnectionAfterReplyPolicy(IMTAPolicy):
            def accept_helo(self, helo_string, message):
                decision = PolicyDecision(False, (552, 'Stupid Spammer'))
                decision._close_connection_after_response = True
                return decision
        self.init(policy=CloseConnectionAfterReplyPolicy())

        self.send('HELO', 'foo.example.com', expected_first_digit=5)
        code, reply_text = self.command_parser.replies[-1]
        self.assert_equals((552, 'Stupid Spammer'), (code, reply_text))
        self.assert_false(self.command_parser.open)

    def test_can_close_connection_before_reply(self):
        class CloseConnectionAfterReplyPolicy(IMTAPolicy):
            def accept_helo(self, helo_string, message):
                decision = PolicyDecision(False)
                decision._close_connection_before_response = True
                return decision
        self.init(policy=CloseConnectionAfterReplyPolicy())

        number_replies_before = len(self.command_parser.replies)
        self.session.handle_input('HELO', 'foo.example.com')
        self.assert_equals(number_replies_before, len(self.command_parser.replies))
        self.assert_false(self.command_parser.open)

    def test_can_close_connection_after_using_default_response(self):
        class CloseConnectionAfterReplyPolicy(IMTAPolicy):
            def accept_helo(self, helo_string, message):
                decision = PolicyDecision(False)
                decision._close_connection_after_response = True
                return decision
        self.init(policy=CloseConnectionAfterReplyPolicy())

        self.send('HELO', 'foo.example.com', expected_first_digit=5)
        code, reply_text = self.command_parser.replies[-1]
        self.assert_equals((550, 'Administrative Prohibition'), (code, reply_text))
        self.assert_false(self.command_parser.open)

    def test_can_close_connection_after_positive_response(self):
        class CloseConnectionAfterPositiveReplyPolicy(IMTAPolicy):
            def accept_helo(self, helo_string, message):
                decision = PolicyDecision(True)
                decision._close_connection_after_response = True
                return decision
        self.init(policy=CloseConnectionAfterPositiveReplyPolicy())

        self.send('HELO', 'foo.example.com')
        self.assert_false(self.command_parser.open)





# -*- coding: UTF-8 -*-
# SPDX-License-Identifier: MIT

from pythonic_testcase import *

from pymta.api import IMTAPolicy
from pymta.command_parser import SMTPCommandParser
from pymta.test_util import BlackholeDeliverer
from pymta.compat import basestring

from tests.util import MockChannel


class CommandParsingTest(PythonicTestCase):

    def setUp(self):
        super(CommandParsingTest, self).setUp()
        self.deliverer = BlackholeDeliverer()
        self.init_command_parser()

    def init_command_parser(self, policy=None):
        self.parser = SMTPCommandParser(MockChannel(), '127.0.0.1', 12345,
                                        self.deliverer, policy=policy)

    def parse_command(self, input):
        return self.parser._parser.parse(input)

    def test_parse_command_without_arguments(self):
        self.assert_equals(('QUIT', None), self.parse_command('QUIT'))
        self.assert_equals(('RSET', None), self.parse_command('RSET'))
        self.assert_equals(('DATA', None), self.parse_command('DATA'))

    def test_parse_helo(self):
        self.assert_equals(('HELO', 'foo.example.com'),
                         self.parse_command('HELO foo.example.com'))
        # This command is syntactically invalid but the validity of specific
        # commands should not be checked in the parser.
        self.assert_equals(('helo', 'foo example.com'),
                         self.parse_command('helo foo example.com'))

    def test_strip_parameters(self):
        self.assert_equals(('HELO', 'foo.example.com'),
                         self.parse_command('HELO   foo.example.com   '))

    def test_parse_commands_with_colons(self):
        self.assert_equals(('MAIL FROM', 'foo@example.com'),
                        self.parse_command('MAIL FROM: foo@example.com'))
        self.assert_equals(('MAIL FROM', 'foo@example.com'),
                        self.parse_command('MAIL FROM:foo@example.com'))
        self.assert_equals(('MAIL FROM', 'foo@example.com'),
                        self.parse_command('MAIL FROM:  foo@example.com   '))
        self.assert_equals(('RCPT TO', 'foo@example.com, bar@example.com'),
                        self.parse_command('RCPT TO:foo@example.com, bar@example.com'))

    def test_parse_auth_plain(self):
        self.assert_equals(('AUTH PLAIN', 'AGZvbwBiYXI='),
                         self.parse_command('AUTH PLAIN AGZvbwBiYXI='))

    def send(self, value):
        if isinstance(value, basestring):
            self.parser.process_new_data(value)
        else:
            for item in value:
                self.parser.process_new_data(item)

    def _send_helo_mail_from_and_rcpt_to(self):
        self.send('HELO foo\r\n')
        self.send('MAIL FROM: foo@example.com\r\n')
        self.send('RCPT TO: bar@example.com\r\n')

    def test_can_switch_to_data_mode(self):
        self._send_helo_mail_from_and_rcpt_to()
        self.send('DATA\r\n')
        self.assert_true(self.parser.is_in_data_mode())

    def test_switch_to_command_mode_when_mail_is_sent(self):
        self._send_helo_mail_from_and_rcpt_to()
        self.send('DATA\r\n')
        self.assert_true(self.parser.is_in_data_mode())
        self.send(['Subject: Foo\r\n\r\n', 'test', '\r\n.\r\n'])
        self.assert_true(self.parser.is_in_command_mode())
        self.assert_equals(1, self.deliverer.received_messages.qsize())

    def received_message(self):
        messages = self.deliverer.received_messages
        self.assert_equals(1, messages.qsize())
        return messages.get()

    def assert_no_messages_received(self):
        self.assert_equals(0, self.deliverer.received_messages.qsize())

    def replies(self):
        return self.parser._channel.replies

    def last_reply(self):
        return self.replies()[-1]

    def test_recognizes_end_of_data_mode_even_if_terminator_is_sent_in_multiple_packages(self):
        self._send_helo_mail_from_and_rcpt_to()
        self.send(['DATA\r\n', '\r\n', '.', '\r\n'])
        self.assert_equals('', self.received_message().msg_data)
        self.assert_true(self.parser.is_in_command_mode())

    def test_recognizes_complete_command_mode_even_if_terminator_is_sent_in_multiple_packages(self):
        self.send(['HELO foo', '\r', '\n'])
        self.assert_equals(2, len(self.replies()))

    def test_supports_transparency_for_lines_starting_with_a_dot(self):
        """SMTP transparency support - see RFC 821, section 4.5.2"""
        self._send_helo_mail_from_and_rcpt_to()
        self.send(['DATA\r\n', '..foo\r\n', '..bar..baz\r\n', '\r\n.\r\n'])
        self.assert_equals('.foo\n.bar..baz\n', self.received_message().msg_data)
        self.assert_true(self.parser.is_in_command_mode())

    def test_big_messages_are_rejected(self):
        """Check that messages which exceed the configured maximum message size
        are rejected. This tests all the code setting the maximum allowed input
        size in the transport layer."""
        class RestrictedSizePolicy(IMTAPolicy):
            def max_message_size(self, peer):
                return 100
        self.init_command_parser(RestrictedSizePolicy())

        self._send_helo_mail_from_and_rcpt_to()
        self.send(['DATA\r\n'])
        self.send(('x'*70 + '\n',) * 1500)
        self.send('\r\n.\r\n')
        self.assert_no_messages_received()
        self.assert_true(self.last_reply().startswith('552 '))



# -*- coding: UTF-8 -*-
# SPDX-License-Identifier: MIT
"""
This module contains some classes which are probably useful for writing unit
tests using pymta:
- MTAThread enables you to run a MTA in a separate thread so that you can test
  interaction with an in-process MTA.
- DebuggingMTA provides a very simple MTA which just collects all incoming
  messages so that you can examine then afterwards.
"""

import random
import socket
import threading
import time

from pythonic_testcase import *

from pymta.api import IMessageDeliverer, IMTAPolicy
from pymta.compat import queue
from pymta.mta import PythonMTA


__all__ = ['BlackholeDeliverer', 'DebuggingMTA', 'MTAThread', 'SMTPTestCase']


class BlackholeDeliverer(IMessageDeliverer):
    """BlackholeDeliverer just stores all received messages in memory in the
    class attribute 'received_messages' (which implements a Queue-like
    interface) so that you can examine the received messages later. """

    received_messages = None

    def __init__(self):
        super(BlackholeDeliverer, self).__init__()
        self.__class__.received_messages = queue.Queue()

    def new_message_accepted(self, msg):
        self.__class__.received_messages.put(msg)


class DebuggingMTA(PythonMTA):
    """DebuggingMTA is a very simple implementation of PythonMTA which just
    collects all incoming messages so that you can examine then afterwards."""

    def __init__(self, *args, **kwargs):
        PythonMTA.__init__(self, *args, **kwargs)
        self.queue = queue.Queue()

    def serve_forever(self):
        return super(DebuggingMTA, self).serve_forever(use_multiprocessing=False)


class MTAThread(threading.Thread):
    """This class runs a PythonMTA in a separate thread which is helpful for
    unit testing.

    Attention: Do not use this class together with multiprocessing!
    http://www.viraj.org/b2evolution/blogs/index.php/2007/02/10/threads_and_fork_a_bad_idea
    """

    def __init__(self, server):
        threading.Thread.__init__(self)
        self.server = server

    def run(self):
        """Create a new thread which runs the server until stop() is called."""
        self.server.serve_forever()

    def stop(self, timeout_seconds=5.0):
        """Stop the mail sink and shut down this thread. timeout_seconds
        specifies how long the caller should wait for the mailsink server to
        close down (default: 5 seconds). If the server did not stop in time, a
        warning message is printed."""
        self.server.shutdown_server()
        threading.Thread.join(self, timeout=timeout_seconds)
        if self.isAlive():
            print("WARNING: Thread still alive. Timeout while waiting for termination!")


class SMTPTestCase(PythonicTestCase):
    """The SMTPTestCase is a unittest.TestCase and provides you with a running
    MTA listening on 'localhost:[8000-40000]' which you can use in your
    tests. No messages will be delivered to the outside world because the MTA
    configured by default uses the BlackholeDeliverer.

    Please make sure that you call the super method for setUp and tearDown."""

    def setUp(self):
        super(SMTPTestCase, self).setUp()
        self.hostname = 'localhost'
        self.listen_port = random.randint(8000, 40000)
        self.mta_thread = None
        self.init_mta()

    def build_mta(self, hostname, listen_port, deliverer, policy_class=None):
        """Return a PythonMTA instance which is configured according to your
        needs."""
        return DebuggingMTA(hostname, listen_port, deliverer,
                            policy_class=policy_class)

    def init_mta(self, policy_class=IMTAPolicy):
        """Starts the MTA in a separate thread with a BlackholeDeliver. This
        method also ensures that the MTA is really listening on the specified
        port."""
        self.stop_mta()
        self.deliverer = BlackholeDeliverer
        self.mta = self.build_mta(self.hostname, self.listen_port,
                                  self.deliverer, policy_class)
        self.mta_thread = MTAThread(self.mta)
        self.mta_thread.start()

        self._try_to_connect_to_mta(self.hostname, self.listen_port)

    def _try_to_connect_to_mta(self, host, port):
        tries = 0
        while tries < 10:
            try:
                sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                sock.connect((host, port))
            except socket.error:
                tries += 1
                time.sleep(0.1)
            else:
                sock.close()
                return
        assert False, 'MTA not reachable'

    def stop_mta(self):
        if self.mta_thread is not None:
            self.mta_thread.stop()
            self.mta_thread = None

    def tearDown(self):
        """Stops the MTA thread."""
        self.stop_mta()
        super(SMTPTestCase, self).tearDown()

    def get_received_messages(self):
        """Return a list of received messages which are stored in the
        BlackholeDeliverer."""
        return self.deliverer.received_messages



# -*- coding: UTF-8 -*-
# SPDX-License-Identifier: MIT

import re
import socket

from pymta.exceptions import SMTPViolationError
from pymta.session import SMTPSession
from pymta.statemachine import StateMachine
from pymta.compat import queue, basestring

__all__ = ['SMTPCommandParser']


class ParserImplementation(object):
    """The SMTPCommandParser needs a connected socket to operate. This is very
    inconvenient for testing therefore all 'interesting' functionality is moved
    in this class which is easily testable."""

    def __init__(self, allowed_commands):
        self._allowed_commands = allowed_commands
        regex_string = '^(%s)(?: |:)\s*(.*)$' % '|'.join(allowed_commands)
        self.parse_regex = re.compile(regex_string, re.IGNORECASE)

    def parse(self, command):
        assert isinstance(command, basestring)
        parameter = None

        match = self.parse_regex.search(command)
        if match is not None:
            command = match.group(1)
            parameter = match.group(2).strip()
        return command, parameter


class SMTPCommandParser(object):
    """This class is a tiny abstraction layer above the real communication
    with the client. It knows about the two basic SMTP modes (sending commands
    vs. sending message data) and can assemble SMTP-like replies (Code-Message)
    in a convenient manner. However all handling of SMTP commands will take
    place in an upper layer.

    The original 'SMTPChannel' class from Python.org handled all communication
    with asynchat, implemented a extremely simple state machine and processed
    the data. Implementing hooks (or adding fine-grained policies) was not
    possible at all in the previous architecture."""

    LINE_TERMINATOR = '\r\n'

    def __init__(self, channel, remote_ip_string, remote_port, deliverer,
                 policy=None, authenticator=None):
        self._channel = channel

        self.data = ''
        self.terminator = self.LINE_TERMINATOR
        self._build_state_machine()

        self.session = SMTPSession(command_parser=self, deliverer=deliverer,
                                   policy=policy, authenticator=authenticator)
        allowed_commands = self.session.get_all_allowed_internal_commands()
        self._parser = ParserImplementation(allowed_commands)
        self.session.new_connection(remote_ip_string, remote_port)
        self._maximum_message_size = None

    def _build_state_machine(self):
        def _command_completed(from_state, to_state, smtp_command, instance):
            self.data = ''

        def _start_receiving_message(from_state, to_state, smtp_command):
            self.terminator = '%s.%s' % (self.LINE_TERMINATOR, self.LINE_TERMINATOR)
            self.data = ''

        def _finished_receiving_message(from_state, to_state, smtp_command):
            self.terminator = self.LINE_TERMINATOR
            self.data = ''

        self.state = StateMachine(initial_state='commands')
        self.state.add('commands', 'commands', 'COMMAND', _command_completed)
        self.state.add('commands', 'data',     'DATA', _start_receiving_message)
        self.state.add('data',     'commands', 'COMMAND', _finished_receiving_message)

    def primary_hostname(self):
        # TODO: This should go into a config object!
        return socket.getfqdn()
    primary_hostname = property(primary_hostname)

    # -------------------------------------------------------------------------
    # Communication helper methods

    def multiline_push(self, code, lines):
        """Send a multi-message to the peer (using the correct SMTP line
        terminators (usually only called from the SMTPSession)."""
        for line in lines[:-1]:
            answer = '%s-%s' % (code, line)
            self.push(answer)
        self.push(code, lines[-1])

    def push(self, code, msg=None):
        """Send a message to the peer (using the correct SMTP line terminators
        (usually only called from the SMTPSession)."""
        if msg is None:
            msg = str(code)
        else:
            msg = '%s %s' % (code, msg)

        if not msg.endswith(self.LINE_TERMINATOR):
            msg += self.LINE_TERMINATOR
        self._channel.write(msg)

    def input_exceeds_limits(self):
        """Called from the underlying transport layer if the client input
        exceeded the configured maximum message size."""
        self.session.input_exceeds_limits()
        self.switch_to_command_mode()

    def is_input_too_big(self):
        if self._maximum_message_size is None:
            return False
        return len(self.data) > self._maximum_message_size

    def set_maximum_message_size(self, max_size):
        """Set the maximum allowed size (in bytes) of a command/message in the
        underlying transport layer so that big messages are not stored in memory
        before they are rejected."""
        self._maximum_message_size = max_size

    def switch_to_command_mode(self):
        """Called from the SMTPSession when a message was received and the
        client is expected to send single commands again."""
        self.state.execute('COMMAND')

    def switch_to_data_mode(self):
        """Called from the SMTPSession when the client should start transfering
        the actual message data."""
        self.state.execute('DATA')

    def _remove_leading_dots_for_smtp_transparency_support(self, input_data):
        """Uses the input data to recover the original payload (includes
        transparency support as specified in RFC 821, Section 4.5.2)."""
        regex = re.compile('^\.\.', re.MULTILINE)
        data_without_transparency_dots = regex.sub('.', input_data)
        return re.sub('\r\n', '\n', data_without_transparency_dots)

    def is_in_command_mode(self):
        assert self.state.state() in ('commands', 'data')
        return self.state.state() == 'commands'

    def is_in_data_mode(self):
        return not self.is_in_command_mode()

    def process_new_data(self, data):
        self.data += data
        if self.is_input_too_big():
            self.session.input_exceeds_limits()
            self.switch_to_command_mode()
            return
        if self.terminator not in self.data:
            return

        # TODO: actually we should not assume that the terminator is at
        # the end of the input string
        input_data_without_terminator = self.data[:-len(self.terminator)]
        # REFACT: add property for this
        if self.is_in_command_mode():
            command, parameter = self._parser.parse(input_data_without_terminator)
            self.session.handle_input(command, parameter)
            self.data = ''
        else:
            msg_data = self._remove_leading_dots_for_smtp_transparency_support(input_data_without_terminator)
            self.session.handle_input('MSGDATA', msg_data)

    def close_when_done(self):
        self._channel.close()


class ClientDisconnectedError(SMTPViolationError):
    """Raised when the SMTP client closed the connection unexpectedly."""
    pass


class WorkerProcess(object):
    """The WorkerProcess handles the real communication. with the client. It
    does not know anything about the SMTP protocol (besides the fact that it is
    a line-based protocol)."""

    def __init__(self, queue, server_socket, deliverer_class, policy_class=None,
                 authenticator_class=None):
        self._queue = queue
        self._server_socket = server_socket
        self._deliverer = self._get_instance_from_class(deliverer_class)
        self._policy = self._get_instance_from_class(policy_class)
        self._authenticator = self._get_instance_from_class(authenticator_class)

        self._connection = None
        self._chatter = None
        self._ignore_write_operations = False

    def _get_instance_from_class(self, class_reference):
        instance = None
        if class_reference is not None:
            instance = class_reference()
        return instance

    def _wait_for_connection(self):
        while True:
            # We want to check periodically if we need to abort
            try:
                connection, remote_address = self._server_socket.accept()
                break
            except socket.timeout:
                try:
                    new_token = self._queue.get_nowait()
                    self._queue.put(new_token)
                    if new_token is None:
                        return None
                except queue.Empty:
                    pass
            except KeyboardInterrupt:
                return None
        connection.settimeout(socket.getdefaulttimeout())
        return connection, remote_address

    def _get_token_with_timeout(self, seconds):
        # wait at max 1 second for the token so that we can abort the whole
        # process in a reasonable time
        token = None
        while True:
            try:
                token = self._queue.get(timeout=seconds)
                break
            except queue.Empty:
                pass
            except KeyboardInterrupt:
                break
        return token

    def run(self):
        token = None
        def have_token():
            return (token is not None)

        try:
            while True:
                token = self._get_token_with_timeout(1)
                if not have_token():
                    break
                assert token is True

                connection_info = self._wait_for_connection()
                self._queue.put(token)
                token = None
                if connection_info is None:
                    break
                self.handle_connection(connection_info)
        finally:
            if have_token():
                # If we possess the token, put it back in the queue so other can
                # continue doing stuff.
                self._queue.put(True)

    def _setup_new_connection(self, connection_info):
        self._connection, (remote_ip_string, remote_port) = connection_info
        self._ignore_write_operations = False
        self._chatter = SMTPCommandParser(self, remote_ip_string, remote_port,
                            self._deliverer, self._policy, self._authenticator)

    def handle_connection(self, connection_info):
        self._setup_new_connection(connection_info)
        try:
            while self.is_connected():
                try:
                    data = self._connection.recv(4096)
                except socket.error:
                    raise ClientDisconnectedError()
                if not data:
                    raise ClientDisconnectedError()
                self._chatter.process_new_data(data.decode('ascii'))
        except ClientDisconnectedError:
            if self.is_connected():
                self.close()

    def is_connected(self):
        return (self._connection is not None)

    def close(self):
        """Closes the connection to the client."""
        if self.is_connected():
            self._connection.close()
            self._connection = None

    def write(self, data):
        """Sends some data to the client."""
        # I don't want to add a separate 'Client disconnected' logic for sending.
        # Therefore I just ignore any writes after the first error - the server
        # won't send that much data anyway. Afterwards the read will detect the
        # broken connection and we quit.
        if self._ignore_write_operations:
            return
        assert self.is_connected()
        try:
            self._connection.send(data.encode('ascii'))
        except socket.error:
            self.close()
            self._ignore_write_operations = True




# -*- coding: UTF-8 -*-
# SPDX-License-Identifier: MIT
"""This package contains classes which from which you can derive your own
classes easily to customize the behavior of your MTA. Everything in here is
considered part of the public API which should be as stable as possible."""

__all__ = ['IAuthenticator', 'IMessageDeliverer', 'IMTAPolicy', 'PolicyDecision',
           'PyMTAException']


class IAuthenticator(object):
    """Authenticators check if the user’s credentials are actually correct. This
    may involve some checking against external subsystems (e.g. a database or a
    LDAP directory)."""

    def authenticate(self, username, password, peer):
        """This method is called after the client issued an AUTH PLAIN command
        and must return a boolean value (True/False)."""
        raise NotImplementedError



class IMessageDeliverer(object):
    """Deliverers take care of the message routing/delivery after a message was
    accepted (e.g. put it in a mailbox file, forward it to another server, ...).
    """

    def new_message_accepted(self, msg):
        """This method is called when a new message was accepted by the server.
        Now the MTA is then in charge of delivering the message to the
        specified recipients. Please note that you can not reject the message
        anymore at this stage (if there are problems you must generate a
        non-delivery report aka bounce).

        There will be one deliverer instance per client connection so
        this method may does not have to be thread-safe. However this method
        may get called multiple times when the client transmits more than one
        message for the same connection."""
        raise NotImplementedError


class PolicyDecision(object):
    def __init__(self, decision=True, reply=None):
        self._decision = decision
        self._reply = reply
        self._close_connection_before_response = False
        self._close_connection_after_response = False

    def close_connection_before_response(self):
        """Return True if the server should close the client connection without
        any further communication."""
        return self._close_connection_before_response

    def close_connection_after_response(self):
        """Return True if the server should close the client connection after it
        sent the given response."""
        return self._close_connection_after_response

    def is_command_acceptable(self):
        return self._decision

    def use_custom_reply(self):
        return (self._reply is not None)

    def get_custom_reply(self):
        if not self.use_custom_reply():
            raise ValueError('No custom reply set.')
        return self._reply


class IMTAPolicy(object):
    """Policies can change with behavior of an MTA dynamically (e.g. don't allow
    relaying unless the client is located within the trusted company network,
    enable authentication only for some connections). In established MTAs like
    Exim and Postfix it's a very important task for every system administrator
    to configure the message acceptance policies which are normally part of the
    configuration file.

    A policy does not change the SMTP implementation itself (the state machine)
    but can send out custom replies to the client. A policy doesn't have to care
    if the commands were given in the correct order (the state machines will
    take care of that). The only thing is that the message object passed into
    many policy methods does not contain all data at certain stages (e.g.
    accept_mail_from can not access the recipients list because that was not
    submitted yet).

    'IMTAPolicy' provides a very permissive policy (all commands are
    accepted) from which you can derive custom policies. Its methods are usually
    named 'accept_<SMTP command name>'.

    Every method in the 'IMTAPolicy' interface can return either a single
    boolean value (True/False) or a tuple. A boolean value specifies if the
    command should be accepted. The caller is responsible for sending the actual
    default replies.

    Alternatively a policy can choose to return a tuple to have more control
    over the reply sent to the client: (decision, (reply code, reply message)).
    The decision is the boolean known from the last paragraph. The reply code
    is an integer which should a be a valid SMTP code. reply message is either a
    basestring with a custom message or an iterable of basestrings (in case a
    a multi-line reply is sent).

    Last but not least a PolicyDecision can be returned which embodies the
    decision as well as (optionally) a custom reply. The reply has the same
    format as described in the paragraph before. The PolicyDecision can ask the
    server to close the connection unconditionally after or even before sending
    the response to the client (in the latter case no response will be sent).
    """

    def accept_new_connection(self, peer):
        """This method is called directly after a new connection is received.
        The  policy can decide if the given peer is allowed to connect to the
        SMTP server. If it declines, the connection will be closed
        immediately."""
        return True

    def max_message_size(self, peer):
        """Return the maximum size (in bytes) for messages from this peer. When
        this method returns an integer, there pymta will check the actual
        message size after the message was received (before the accept_msgdata
        method is called) and will respond with the appropriate error message if
        necessary.
        If you return None, no size limit will be enforced by pymta (however you
        can always reject a message using accept_msgdata()."""
        return None

    def ehlo_lines(self, peer):
        """Return an iterable for SMTP extensions to advertise after EHLO.
        By default support for SMTP SIZE extension will be announced if you set
        a max message size."""
        max_size = self.max_message_size(peer)
        if max_size != None:
            return ('SIZE %d' % max_size,)
        return ()

    def accept_helo(self, helo_string, message):
        """Decides if the HELO command with the given helo_name should be
        accepted."""
        return True

    def accept_ehlo(self, ehlo_string, message):
        """Decides if the EHLO command with the given helo_name should be
        accepted."""
        return True

    def accept_auth_plain(self, username, password, message):
        """Decides if AUTH plain should be allowed for this client. Please note
        that username and password are not verified before, the authenticator
        will check them after the policy allowed this command.

        The method must not return a response by itself in case it accepts the
        AUTH PLAIN command!"""
        return True

    def accept_from(self, sender, message):
        "Decides if the sender of this message (MAIL FROM) should be accepted."
        return True

    def accept_rcpt_to(self, new_recipient, message):
        """Decides if recipient of this message (RCPT TO) should be accepted.
        If a message should be delivered to multiple recipients this method is
        called for every recipient."""
        return True

    def accept_data(self, message):
        """Decides if we allow the client to start a message transfer (the
        actual message contents will be transferred after this method allowed
        it)."""
        return True

    def accept_msgdata(self, msgdata, message):
        """This method actually matches no real SMTP command. It is called
        after a message was transferred completely and this is the last check
        before the SMTP server takes the responsibility of transferring it to
        the recipients."""
        return True


class PyMTAException(Exception):
    """Base class for all exceptions used in pymta."""
    pass







# -*- coding: UTF-8 -*-
# SPDX-License-Identifier: MIT

from pymta.api import IMTAPolicy
from pymta.compat import b64encode

from tests.util import CommandParserTestCase, DummyAuthenticator


class BasicMessageSendingTest(CommandParserTestCase):

    def _check_last_code(self, expected_code):
        code, reply_text = self.command_parser.replies[-1]
        self.assert_equals(expected_code, code)

    def test_new_connection(self):
        self.assert_equals(1, len(self.command_parser.replies))
        self._check_last_code(220)
        code, reply_text = self.command_parser.replies[-1]
        self.assert_equals('localhost Hello 127.0.0.1', reply_text)
        self.close_connection()

    def test_noop_does_nothing(self):
        self.send('noop')
        self.close_connection()

    def test_send_helo(self):
        self.send('helo', 'foo.example.com')
        self.assert_equals(2, len(self.command_parser.replies))
        self._check_last_code(250)
        code, reply_text = self.command_parser.replies[-1]
        self.assert_equals('localhost', reply_text)
        self.close_connection()

    def test_reject_duplicated_helo(self):
        self.send('helo', 'foo.example.com')
        code, reply_text = self.send('helo', 'foo.example.com',
                                      expected_first_digit=5)
        self.assert_equals(503, code)
        expected_message = 'Command "helo" is not allowed here'
        self.assert_true(reply_text.startswith(expected_message), reply_text)
        self.close_connection()

    def test_helo_without_hostname_is_rejected(self):
        self.send('helo', expected_first_digit=5)
        # But we must be able to send the right command here (state machine must
        # not change)
        self.send('helo', 'foo')

    def test_helo_with_invalid_arguments_is_rejected(self):
        expect_invalid = lambda data: self.assert_equals(501, self.send('helo', data, expected_first_digit=5)[0])
        expect_invalid('')
        expect_invalid('  ')
        expect_invalid(None)
        # Even if we don't enforce that the helo parameter must be a valid host
        # name (as required as per RFC 2821), at least there should be only
        # one parameter.
        expect_invalid('foo bar')

    def test_helo_can_send_ipv4_address_in_brackets(self):
        # smtplib in Python 2.6.2 does this at least...
        self.send('helo', '[127.0.0.1]')

    def test_invalid_commands_are_recognized(self):
        self.session.handle_input('invalid')
        self.assert_equals(2, len(self.command_parser.replies))
        self._check_last_code(500)
        code, reply_text = self.command_parser.replies[-1]
        self.assert_equals('unrecognized command "invalid"', reply_text)
        self.close_connection()

    def _send_mail(self, rfc822_msg):
        self.send('MAIL FROM', 'foo@example.com')
        self.send('RCPT TO', 'bar@example.com')
        self.send('DATA', expected_first_digit=3)
        self.send('MSGDATA', rfc822_msg)

    def test_send_simple_mail(self):
        rfc822_msg = 'Subject: Test\n\nJust testing...\n'
        self.send('HELO', 'foo.example.com')
        self._send_mail(rfc822_msg)
        self.close_connection()

        received_messages = self.deliverer.received_messages
        self.assert_equals(1, received_messages.qsize())
        msg = received_messages.get()
        self.assert_equals('127.0.0.1', msg.peer.remote_ip)
        self.assert_equals(4567, msg.peer.remote_port)

        self.assert_equals('foo.example.com', msg.smtp_helo)
        self.assert_equals('foo@example.com', msg.smtp_from)
        self.assert_equals(['bar@example.com'], msg.smtp_to)
        self.assert_equals(rfc822_msg, msg.msg_data)

    def test_help_is_supported(self):
        code, reply_text = self.send('HELP')
        self.assert_equals(214, code)
        supported_commands = set(reply_text[1].split(' '))
        expected_commands = set(['AUTH', 'DATA', 'EHLO', 'HELO', 'HELP', 'MAIL',
                                 'NOOP', 'QUIT', 'RCPT', 'RSET'])
        self.assert_equals(expected_commands, supported_commands)

    def test_support_for_rset(self):
        self.send('HELO', 'foo.example.com')
        self.send('MAIL FROM', 'foo@example.com')
        self.send('RSET')
        self.send('MAIL FROM', 'bar@example.com')

    def test_send_ehlo_without_authenticator(self):
        self.send('EHLO', 'foo.example.com')
        self.assert_equals(2, len(self.command_parser.replies))
        code, reply_text = self.command_parser.replies[-1]
        self.assert_equals(250, code)
        self.assert_equals(set(('localhost', 'HELP')), set(reply_text))

    def test_ehlo_without_hostname_is_rejected(self):
        self.send('EHLO', expected_first_digit=5)

    def test_ehlo_with_invalid_arguments_is_rejected(self):
        expect_invalid = lambda data: self.assert_equals(501, self.send('ehlo', data, expected_first_digit=5)[0])
        expect_invalid('')
        expect_invalid('  ')
        expect_invalid(None)
        expect_invalid('foo bar')

    def test_auth_plain_without_authenticator_is_rejected(self):
        self.send('EHLO', 'foo.example.com')
        base64_credentials = b64encode('\x00foo\x00foo')
        self.send('AUTH PLAIN', base64_credentials, expected_first_digit=5)
        self.assert_equals(3, len(self.command_parser.replies))
        self._check_last_code(535)

    def test_authenticator_advertises_auth_plain_support(self):
        self.session._authenticator = DummyAuthenticator()

        self.send('EHLO', 'foo.example.com')
        code, reply_texts = self.command_parser.replies[-1]
        self.assert_true('AUTH PLAIN' in reply_texts)

    def test_auth_plain_with_username_and_password_is_accepted(self):
        self.session._authenticator = DummyAuthenticator()

        self.send('EHLO', 'foo.example.com')
        self.send('AUTH PLAIN', b64encode('\x00foo\x00foo'))
        self.assert_equals(3, len(self.command_parser.replies))
        self._check_last_code(235)
        code, reply_text = self.command_parser.replies[-1]
        self.assert_equals('Authentication successful', reply_text)

    def test_auth_plain_with_authzid_username_and_password_is_accepted(self):
        self.session._authenticator = DummyAuthenticator()

        self.send('EHLO', 'foo.example.com')
        # RFC 4616 defines SASL PLAIN in the form
        # [authzid] \x00 authcid \x00 passwd
        # smtplib in Python 2.3 will send an additional authzid (which is equal
        # to authcid)
        self.send('AUTH PLAIN', b64encode('ignored\x00foo\x00foo'))
        self.assert_equals(3, len(self.command_parser.replies))
        self._check_last_code(235)
        code, reply_text = self.command_parser.replies[-1]
        self.assert_equals('Authentication successful', reply_text)

    def test_auth_plain_with_bad_credentials_is_accepted(self):
        self.session._authenticator = DummyAuthenticator()

        self.send('EHLO', 'foo.example.com')
        base64_credentials = b64encode('\x00foo\x00bar')
        self.send('AUTH PLAIN', base64_credentials, expected_first_digit=5)
        self._check_last_code(535)

    def test_auth_plain_with_bad_base64_is_rejected(self):
        self.session._authenticator = DummyAuthenticator()

        self.send('EHLO', 'foo.example.com')
        self.send('AUTH PLAIN', 'foo', expected_first_digit=5)
        self.assert_equals(3, len(self.command_parser.replies))
        self._check_last_code(501)

    def test_auth_plain_with_bad_format_is_rejected(self):
        self.session._authenticator = DummyAuthenticator()

        self.send('EHLO', 'foo.example.com')
        base64_credentials = b64encode('\x00foo')
        self.send('AUTH PLAIN', base64_credentials, expected_first_digit=5)
        self.assert_equals(3, len(self.command_parser.replies))
        self._check_last_code(501)

    def test_size_restrictions_are_announced_in_ehlo_reply(self):
        class RestrictedSizePolicy(IMTAPolicy):
            def max_message_size(self, peer):
                return 100
        self.session._policy = RestrictedSizePolicy()

        self.send('EHLO', 'foo.example.com')
        code, reply_texts = self.command_parser.replies[-1]
        self.assert_true('SIZE 100' in reply_texts)

    def test_early_rejection_if_size_verb_indicates_big_message(self):
        class RestrictedSizePolicy(IMTAPolicy):
            def max_message_size(self, peer):
                return 100
        self.session._policy = RestrictedSizePolicy()

        self.send('EHLO', 'foo.example.com')
        self.send('MAIL FROM', '<foo@example.com>   size=106530  ',
                  expected_first_digit=5)
        self._check_last_code(552)

    def test_reject_verbs_for_plain_smtp(self):
        """Test that SMTP extension verbs are rejected when the connection uses
        plain SMTP."""
        self.send('HELO', 'foo.example.com')
        self.send('MAIL FROM', '<foo@example.com>   size=106530  ',
                  expected_first_digit=5)
        code, reply_text = self.command_parser.replies[-1]
        self.assert_equals(501, code)
        self.assert_equals('No SMTP extensions allowed for plain SMTP.', reply_text)

    def test_can_still_use_esmtp_after_first_mail(self):
        self.send('EHLO', 'foo.example.com')
        self._send_mail('Subject: First Message\n\nJust testing...\n')
        self.send('MAIL FROM', '<foo@example.com>   size=106530  ')




# -*- coding: UTF-8 -*-
# SPDX-License-Identifier: MIT


__all__ = ['Message', 'Peer']


class Message(object):
    def __init__(self, peer, smtp_helo=None, smtp_from=None, smtp_to=None,
                 msg_data=None, username=None):
        self.peer = peer
        self.smtp_helo = smtp_helo
        self.smtp_from = smtp_from
        if smtp_to == None:
            smtp_to = []
        elif not isinstance(smtp_to, (list, tuple)):
            smtp_to = [smtp_to]
        self.smtp_to = smtp_to
        self.msg_data = msg_data
        self.username = username



class Peer(object):
    def __init__(self, remote_ip, remote_port):
        self.remote_ip = remote_ip
        self.remote_port = remote_port

    def __repr__(self):
        return '%s(%s, %s)' % (self.__class__.__name__, self.remote_ip,
                               self.remote_port)





# -*- coding: UTF-8 -*-
# SPDX-License-Identifier: MIT


from pycerberus.errors import InvalidDataError
from pycerberus.validators import StringValidator
from pythonic_testcase import *

from pymta.validation import AuthPlainSchema, MailFromSchema, SMTPCommandArgumentsSchema
from pymta.compat import b64encode


class CommandWithoutParametersTest(PythonicTestCase):

    def schema(self):
        return SMTPCommandArgumentsSchema()

    def test_accept_command_without_parameters(self):
        self.assert_equals({}, self.schema().process(''))

    def test_bails_out_if_additional_parameters_are_passed(self):
        e = self.assert_raises(InvalidDataError, lambda: self.schema().process('fnord'))
        self.assert_equals("Syntactically invalid argument(s) 'fnord'", e.msg())


class CommandWithSingleParameterTest(PythonicTestCase):

    def schema(self):
        schema = SMTPCommandArgumentsSchema()
        schema.set_internal_state_freeze(False)
        return schema

    def test_accepts_one_parameter(self):
        schema = self.schema()
        schema.add('parameter', StringValidator)
        schema.set_parameter_order(('parameter',))
        self.assert_equals({'parameter': 'fnord'}, schema.process('fnord'))

    def test_bails_out_if_no_parameter_is_passed(self):
        schema = self.schema()
        schema.add('parameter', StringValidator)
        self.assert_raises(InvalidDataError, lambda: schema.process(''))

    def test_bails_out_if_more_than_one_parameter_is_passed(self):
        schema = self.schema()
        schema.add('parameter', StringValidator)
        self.assert_raises(InvalidDataError, lambda: schema.process('fnord extra'))

    def test_can_use_custom_name_for_parameters(self):
        schema = self.schema()
        schema.add('helo', StringValidator)
        schema.set_parameter_order(('helo', ))
        self.assert_equals({'helo': 'localhost'}, schema.process('localhost'))

    def test_can_specify_parameter_order_declaratively(self):
        class SchemaWithOrderedParameters(SMTPCommandArgumentsSchema):
            foo = StringValidator()
            bar = StringValidator()
            parameter_order = ('foo', 'bar')

        schema = SchemaWithOrderedParameters()
        self.assert_equals({'foo': 'baz', 'bar': 'qux'}, schema.process('baz qux'))


class MailFromSchemaTest(PythonicTestCase):

    def schema(self):
        return MailFromSchema()

    def process(self, input_string, esmtp=None):
        context = {}
        if esmtp is not None:
            context['esmtp'] = esmtp
        return self.schema().process(input_string, context=context)

    # --------------------------------------------------------------------------
    # validating the email address

    def test_accept_plain_email_address(self):
        self.assert_dict_contains({'email': 'foo@example.com'}, self.process('foo@example.com'))

    # --------------------------------------------------------------------------
    # SMTP extensions

    def test_reject_extensions_for_plain_smtp(self):
        input_command = '<foo@example.com> SIZE=1000'
        call = lambda: self.process(input_command, esmtp=False)
        e = self.assert_raises(InvalidDataError, call)
        self.assert_equals('No SMTP extensions allowed for plain SMTP.', e.msg())

    def test_can_parse_extensions(self):
        schema = self.schema()
        schema.add('body', StringValidator())
        input_command = '<foo@example.com> BODY=BINARYMIME'
        self.assert_dict_contains({'email': 'foo@example.com', 'body': 'BINARYMIME'},
                                  schema.process(input_command, context={'esmtp': True}))

    def test_ignores_whitespace_surrounding_extensions(self):
        schema = self.schema()
        schema.add('body', StringValidator())
        input_command = '<foo@example.com>   BODY=BINARYMIME  '
        self.assert_dict_contains({'email': 'foo@example.com', 'body': 'BINARYMIME'},
                                  schema.process(input_command, context={'esmtp': True}))

    def test_treats_extensions_as_case_insensitive(self):
        schema = self.schema()
        schema.add('body', StringValidator())
        input_command = '<foo@example.com> bOdY=BINARYMIME'
        self.assert_dict_contains({'email': 'foo@example.com', 'body': 'BINARYMIME'},
                                  schema.process(input_command, context={'esmtp': True}))

    def test_present_meaningful_error_message_for_unknown_arguments(self):
        input_command = 'foo@example.com foo bar'
        call = lambda: self.process(input_command, esmtp=True)
        e = self.assert_raises(InvalidDataError, call)
        self.assert_equals('Invalid arguments: \'foo bar\'', e.msg())

    def test_present_meaningful_error_message_for_unknown_extensions(self):
        input_command = 'foo@example.com invalid=fnord'
        call = lambda: self.process(input_command, esmtp=True)
        e = self.assert_raises(InvalidDataError, call)
        self.assert_equals('Invalid extension: \'invalid=fnord\'', e.msg())


    # ----------------------------------------------------------------------------
    # Tests for validation of specific extensions

    # --------------------------------------------------------------------------
    # SMTP SIZE extension

    def test_can_extract_size_parameter_if_esmtp_is_enabled(self):
        input_command = 'foo@example.com SIZE=1000'
        self.assert_dict_contains({'size': 1000}, self.process(input_command, esmtp=True))

    def test_size_parameter_is_not_mandatory_even_when_using_esmtp(self):
        parameters = self.process('foo@example.com', esmtp=True)
        self.assert_dict_contains({'email': 'foo@example.com'}, parameters)

    def test_reject_size_below_zero(self):
        input_command = 'foo@example.com SIZE=-1234'
        call = lambda: self.process(input_command, esmtp=True)
        e = self.assert_raises(InvalidDataError, call)
        self.assert_equals('Invalid size: Must be 1 or greater.', e.msg())

    def test_reject_non_numeric_size_parameter(self):
        input_command = 'foo@example.com SIZE=fnord'
        self.assert_raises(InvalidDataError, lambda: self.process(input_command, esmtp=True))


class AuthPlainSchemaTest(PythonicTestCase):

    def schema(self):
        return AuthPlainSchema()

    def process(self, input_string, esmtp=None):
        context = {}
        if esmtp is not None:
            context['esmtp'] = esmtp
        return self.schema().process(input_string, context=context)

    def test_can_extract_base64_decoded_string(self):
        expected_parameters = dict(username='foo', password='foo ', authzid=None)
        parameters = self.schema().process(self.base64('\x00foo\x00foo '))
        self.assert_equals(expected_parameters, parameters)

    def base64(self, value):
        return b64encode(value).strip()

    def assert_bad_input(self, input):
        e = self.assert_raises(InvalidDataError, lambda: self.schema().process(input))
        return e

    def test_reject_more_than_one_parameter(self):
        input = self.base64('\x00foo\x00foo') + ' ' + self.base64('\x00foo\x00foo')
        self.assert_bad_input(input)

    def test_rejects_bad_base64(self):
        e = self.assert_bad_input('invalid')
        self.assert_equals('Garbled data sent', e.msg())

    def test_rejects_invalid_format(self):
        e = self.assert_bad_input(b64encode('foobar'))
        self.assert_equals('Garbled data sent', e.msg())





# -*- coding: UTF-8 -*-
# SPDX-License-Identifier: MIT

from pythonic_testcase import *

from pymta import SMTPSession
from pymta.api import IAuthenticator
from pymta.test_util import BlackholeDeliverer


class MockCommandParser(object):
    primary_hostname = 'localhost'

    def __init__(self):
        self.replies = []
        self.messages = []
        self.open = True

    def set_maximum_message_size(self, max_size):
        pass

    def push(self, code, text):
        assert self.open
        self.replies.append((code, text))

    def multiline_push(self, code, lines):
        assert self.open
        self.replies.append((code, lines))

    def close_when_done(self):
        assert self.open
        self.open = False

    def new_message_received(self, msg):
        self.messages.append(msg)

    def switch_to_command_mode(self):
        pass

    def switch_to_data_mode(self):
        pass


class MockChannel(object):
    def __init__(self):
        self.replies = []

    def write(self, data):
        self.replies.append(data)

    def close(self):
        pass


class DummyAuthenticator(IAuthenticator):
    def authenticate(self, username, password, peer):
        return username == password


class CommandParserTestCase(PythonicTestCase):

    def setUp(self, policy=None):
        super(CommandParserTestCase, self).setUp()
        self.init(policy=policy)

    def tearDown(self):
        if self.command_parser.open:
            self.close_connection()
        super(CommandParserTestCase, self).tearDown()

    def init(self, policy=None, authenticator=None):
        self.command_parser = MockCommandParser()
        self.deliverer = BlackholeDeliverer()
        self.session = SMTPSession(command_parser=self.command_parser,
                                   deliverer=self.deliverer,
                                   policy=policy, authenticator=authenticator)
        self.session.new_connection('127.0.0.1', 4567)

    def check_reply_code(self, code, reply_text, expected_first_digit):
        first_code_digit = int(str(code)[0])
        smtp_reply = "%s %s" % (code, reply_text)
        if expected_first_digit is not None:
            self.assert_equals(expected_first_digit, first_code_digit, smtp_reply)

    def send(self, command, data=None, expected_first_digit=2):
        number_replies_before = len(self.command_parser.replies)
        self.session.handle_input(command, data)
        self.assert_equals(number_replies_before + 1, len(self.command_parser.replies))
        code, reply_text = self.command_parser.replies[-1]
        self.check_reply_code(code, reply_text, expected_first_digit=expected_first_digit)
        return (code, reply_text)

    def close_connection(self):
        self.send('quit', expected_first_digit=2)
        code, reply_text = self.command_parser.replies[-1]
        self.assert_equals(221, code)
        self.assert_equals('localhost closing connection', reply_text)
        self.assert_false(self.command_parser.open)





# -*- coding: UTF-8 -*-
# SPDX-License-Identifier: MIT

from pythonic_testcase import *

from pymta.statemachine import StateMachine, StateMachineDefinitionError, \
    StateMachineError


class StateMachineTest(PythonicTestCase):

    def setUp(self):
        super(StateMachineTest, self).setUp()
        self.state = StateMachine(initial_state='new')

    def test_can_initialize_statemachine(self):
        StateMachine(initial_state='foo')

    # --- adding states ------------------------------------------------------

    def test_can_add_states(self):
        self.state.add('new', 'processed', 'process')
        self.state.add('new', 'new', 'noop')

    def test_raise_exception_if_duplicate_action_is_defined(self):
        self.state.add('new', 'processed', 'process')
        with assert_raises(StateMachineDefinitionError):
            self.state.add('new', 'new', 'process')

    # --- introspection ------------------------------------------------------

    def test_can_ask_for_current_state(self):
        state = StateMachine(initial_state='foo')
        state.add('foo', 'foo', 'noop')
        self.assert_equals('foo', state.state())
        self.assert_false(state.is_impossible_state())

    def test_no_state_if_initial_state_not_available(self):
        state = StateMachine(initial_state='invalid')
        self.assert_none(state.state())
        self.assert_true(state.is_impossible_state())

    def test_can_ask_for_all_known_actions(self):
        self.state.add('new', 'new', 'noop')
        self.state.add('new', 'processed', 'process')
        self.state.add('processed', 'new', 'rework')
        self.assert_equals(set(('noop', 'process', 'rework')), self.state.known_actions())

    def test_can_ask_for_all_currently_allowed_actions(self):
        self.state.add('new', 'new', 'noop')
        self.state.add('new', 'processed', 'process')
        self.state.add('processed', 'new', 'rework')

        self.assert_equals(set(('noop', 'process')), self.state.allowed_actions())
        self.state.set_state('processed')
        self.assert_equals(set(('rework',)), self.state.allowed_actions())

    def test_can_ask_for_all_known_states(self):
        self.assert_equals(set(), self.state.known_states())
        self.state.add('new', 'processed', 'process')
        self.state.add('processed', 'done', 'finalize')
        self.assert_equals(set(('new', 'processed', 'done')), self.state.known_states())

    def test_can_ask_for_all_non_final_states(self):
        self.assert_equals(set(), self.state.known_non_final_states())
        self.state.add('new', 'processed', 'process')
        self.state.add('processed', 'done', 'finalize')
        self.assert_equals(set(('new', 'processed')), self.state.known_non_final_states())

    # --- handling states ----------------------------------------------------

    def test_can_not_set_state_to_invalid_state(self):
        with assert_raises(StateMachineError):
            self.state.set_state('invalid')

    # --- executing ----------------------------------------------------------

    def test_can_execute_states(self):
        self.state.add('new', 'processed', 'process')
        self.state.execute('process')
        self.assert_equals('processed', self.state.state())

    def test_handler_is_called_for_state_transition(self):
        self._transition = None
        def handler(from_state, to_state, action_name):
            self._transition = (from_state, to_state, action_name)

        self.state.add('new', 'new', 'noop', handler)
        self.state.execute('noop')
        self.assert_equals('new', self.state.state())
        self.assert_equals(('new', 'new', 'noop'), self._transition)

    def test_raise_exception_for_invalid_action(self):
        self.state.add('new', 'processed', 'process')
        with assert_raises(StateMachineError):
            self.state.execute('invalid')

        self.state.add('processed', 'new', 'rework')
        with assert_raises(StateMachineError):
            self.state.execute('rework')
        self.state.execute('process')
        with assert_raises(StateMachineError):
            self.state.execute('process')
        self.state.execute('rework')

    def test_raise_exception_if_in_impossible_state(self):
        state = StateMachine(initial_state='invalid')
        state.add('new', 'processed', 'process')
        with assert_raises(StateMachineError):
            self.state.execute('process')

    def test_raise_exception_if_no_outgoing_transition_defined_when_executing(self):
        self.state.add('new', 'processed', 'process')
        self.state.set_state('processed')
        with assert_raises(StateMachineError):
            self.state.execute('rework')

    # --- transition with operations and conditions --------------------------

    def test_can_add_transition_with_additional_operation(self):
        self.state.add('new', 'processed', 'process', operations=('set_foo',))

    def test_can_tell_if_flag_is_set(self):
        self.assert_false(self.state.is_set(None))
        self.assert_false(self.state.is_set('foo'))

    def test_transition_can_also_set_flags(self):
        self.state.add('new', 'processed', 'process', operations=('set_foo',))
        self.assert_false(self.state.is_set('foo'))

        self.state.execute('process')
        self.assert_true(self.state.is_set('foo'))

    def test_can_add_conditional_transition(self):
        self.state.add('new', 'authenticated', 'authenticate', condition='if_tls')

    def test_allowed_actions_obeys_condition(self):
        self.state.add('new', 'new', 'use_tls', operations=('set_tls',))
        self.state.add('new', 'authenticated', 'authenticate', condition='if_tls')
        self.assert_equals(set(('use_tls',)), self.state.allowed_actions())
        self.state.execute('use_tls')
        self.assert_equals(set(('use_tls', 'authenticate')), self.state.allowed_actions())

    def test_conditional_transition_is_only_executed_if_flag_is_true(self):
        self.state.add('new', 'new', 'use_tls', operations=('set_tls',))
        self.state.add('new', 'authenticated', 'authenticate', condition='if_tls')
        self.assert_equals('new', self.state.state())
        with assert_raises(StateMachineError):
            self.state.execute('authenticate')
        self.state.execute('use_tls')
        self.assert_true(self.state.is_set('tls'))
        self.state.execute('authenticate')

    def test_can_also_specify_negative_flag_checks_for_transitions(self):
        self.state.add('new', 'new', 'use_tls', operations=('set_tls',), condition='if_not_tls')
        self.state.add('new', 'authenticated', 'authenticate', condition='if_tls')

        with assert_raises(StateMachineError):
            self.state.execute('authenticate')

        self.state.execute('use_tls')
        with assert_raises(StateMachineError):
            self.state.execute('use_tls')
        self.state.execute('authenticate')




# -*- coding: UTF-8 -*-
# SPDX-License-Identifier: MIT

import re

from pycerberus.i18n import _
from pycerberus.schemas import PositionalArgumentsParsingSchema
from pycerberus.validators import EmailAddressValidator, IntegerValidator, StringValidator

from pymta.compat import b64decode


__all__ = ['HeloSchema', 'MailFromSchema', 'RcptToSchema', 'SMTPCommandArgumentsSchema']

# ------------------------------------------------------------------------------
# General infrastructure

class SMTPCommandArgumentsSchema(PositionalArgumentsParsingSchema):

    def messages(self):
        return {'additional_item': _("Syntactically invalid argument(s) '%(additional_item)s'")}

    def separator_pattern(self):
        return '\s+'


class SMTPEmailValidator(EmailAddressValidator):

    def messages(self):
        return {'unbalanced_quotes': _('Invalid email address format - use balanced angle brackets.')}

    def convert(self, value, context):
        string_value = super(SMTPEmailValidator, self).convert(value, context)
        if string_value.startswith('<') or string_value.endswith('>'):
            match = re.search('^<(.+)>$', string_value)
            if match is None:
                self.raise_error('unbalanced_quotes', string_value, context)
            string_value = match.group(1)
        return string_value

# ------------------------------------------------------------------------------
# MAIL FROM

class SizeExtensionValidator(IntegerValidator):

    def __init__(self, *args, **kwargs):
        kwargs.update({'required': False, 'min': 1})
        super(SizeExtensionValidator, self).__init__(*args, **kwargs)

    def messages(self):
        return {'too_low': _('Invalid size: Must be %(min)s or greater.')}


class MailFromSchema(SMTPCommandArgumentsSchema):
    email = SMTPEmailValidator()
    size  = SizeExtensionValidator()

    parameter_order = ('email',)

    def messages(self):
        return {
            'invalid_extension': _('Invalid extension: %(smtp_extension)s'),
            'invalid_smtp_arguments': _('Invalid arguments: %(smtp_arguments)s'),
            'no_extensions': _('No SMTP extensions allowed for plain SMTP.'),
        }

    # --------------------------------------------------------------------------
    # special implementations for SMTP extensions

    def uses_esmtp(self, context):
        return context.get('esmtp', False)

    def _assert_all_options_have_a_value_assigned(self, key_value_pairs, input_string, context):
        for option in key_value_pairs:
            if len(option) == 2:
                continue
            value = ''.join(option)
            self.raise_error('invalid_smtp_arguments', value, context, smtp_arguments=repr(input_string))

    def _assert_only_known_extensions(self, key_value_pairs, input_string, context):
        for key, value in key_value_pairs:
            if key.lower() in self.fieldvalidators():
                continue
            value = '='.join((key, value))
            self.raise_error('invalid_extension', value, context, smtp_extension=repr(input_string))

    def _validate_extension_arguments(self, key_value_pairs, input_string, context):
        self._assert_all_options_have_a_value_assigned(key_value_pairs, input_string, context)
        self._assert_only_known_extensions(key_value_pairs, input_string, context)

    def aggregate_values(self, parameter_names, arguments, context):
        if len(arguments) <= 1:
            return super(MailFromSchema, self).aggregate_values(parameter_names, arguments, context)
        key_value_pairs = [re.split('=', option, 1) for option in arguments[1:]]

        self._validate_extension_arguments(key_value_pairs, ' '.join(arguments[1:]), context)
        lower_case_key_value_pairs = [(item[0].lower(), item[1]) for item in key_value_pairs]
        options = dict(lower_case_key_value_pairs)

        parameter_names = parameter_names + list(options)
        arguments = [arguments[0]] + list(options.values())
        return parameter_names, arguments

    def _process_fields(self, fields, result, context=None):
        is_pycerberus06 = (context is None)
        if is_pycerberus06:
            context = result
            result = None

        if len(fields) > 1 and not self.uses_esmtp(context):
            self.raise_error('no_extensions', '', context)

        if is_pycerberus06:
            return super(MailFromSchema, self)._process_fields(fields, context)
        else:
            return super(MailFromSchema, self)._process_fields(fields, result, context)

# ------------------------------------------------------------------------------

class HeloSchema(SMTPCommandArgumentsSchema):
    helo = StringValidator(strip=True)

    parameter_order = ('helo',)


class RcptToSchema(SMTPCommandArgumentsSchema):
    email = SMTPEmailValidator()

    parameter_order = ('email',)

# ------------------------------------------------------------------------------
# AUTH PLAIN

class AuthPlainSchema(SMTPCommandArgumentsSchema):
    authzid  = StringValidator(required=False, default=None)
    username = StringValidator()
    password = StringValidator()

    parameter_order = ('authzid', 'username', 'password')

    def messages(self):
        return {
            'invalid_base64': _('Garbled data sent'),
            'invalid_format': _('Garbled data sent'),
        }

    # --------------------------------------------------------------------------
    # special implementations for SMTP extensions

    def _decode_base64(self, value, context):
        try:
            return b64decode(value)
        except:
            self.raise_error('invalid_base64', value, context)

    def split_parameters(self, value, context):
        match = re.search(('=\s(.+)$'), value.strip())
        if match is not None:
            self.raise_error('additional_item', value, context, additional_item=repr(match.group(1)))
        decoded_parameters = self._decode_base64(value, context)
        match = re.search('^([^\x00]*)\x00([^\x00]*)\x00([^\x00]*)$', decoded_parameters)
        if not match:
            self.raise_error('invalid_format', value, context)
        parameters = list(match.groups())
        if parameters[0] == '':
            parameters[0] = None
        return parameters




# -*- coding: UTF-8 -*-
# SPDX-License-Identifier: MIT

import sys

from pycerberus import InvalidDataError

from pymta.compat import basestring
from pymta.exceptions import InvalidParametersError, SMTPViolationError
from pymta.model import Message, Peer
from pymta.statemachine import StateMachine, StateMachineError
from pymta.validation import AuthPlainSchema, HeloSchema, MailFromSchema, \
    RcptToSchema, SMTPCommandArgumentsSchema


__all__ = ['SMTPSession']



class PolicyDenial(SMTPViolationError):
    pass


class SMTPSession(object):
    """The SMTPSession processes all input data which were extracted from
    sockets previously. The idea behind this is that this class only knows about
    different SMTP commands and does not have to know things like command mode
    and data mode.

    The protocol parser will create a new session instance for every new
    connection so this class does not have to be thread-safe.
    """

    def __init__(self, command_parser, deliverer, policy=None,
                 authenticator=None):
        self._command_parser = command_parser
        self._deliverer = deliverer
        self._policy = policy
        self._authenticator = authenticator

        self._command_arguments = None
        self._close_connection_after_response = False
        self._is_connected = True
        self._message = None

        self._build_state_machine()

    # -------------------------------------------------------------------------
    # State machine building

    def _add_state(self, from_state, to_state, smtp_command, **kwargs):
        handler_function = self._dispatch_commands
        self.state.add(from_state, smtp_command, to_state, handler_function, **kwargs)

    def _get_all_commands(self, including_quit=False):
        commands = set()
        for actions in self.state._transitions.values():
            for command_name, transition in actions.items():
                target_state = transition[0]
                if target_state in ['new']:
                    continue
                if including_quit or (target_state != 'finished'):
                    commands.add(command_name)
        return commands

    def get_all_allowed_internal_commands(self):
        """Returns an iterable which includes all allowed commands. This does
        not mean that a specific command from the result is executable right now
        in this session state (or that it can be executed at all in this
        connection).

        Please note that the returned values are /internal/ commands, not SMTP
        commands (use get_all_allowed_smtp_commands for that) so there will be
        'MAIL FROM' instead of 'MAIL'."""
        states = set()
        for command_name in self._get_all_commands(including_quit=True):
            if command_name not in ['GREET', 'MSGDATA']:
                states.add(command_name)
        return states

    def get_all_allowed_smtp_commands(self):
        states = set()
        for command_name in self.get_all_allowed_internal_commands():
            command_name = command_name.split(' ')[0]
            states.add(command_name)
        return states

    def _add_rset_transitions(self):
        for state_name in self.state.known_non_final_states():
            if state_name == 'new':
                self._add_state(state_name, 'RSET',  state_name)
            else:
                self._add_state(state_name, 'RSET',  'initialized')

    def _add_help_noop_and_quit_transitions(self):
        """HELP, NOOP and QUIT should be possible from everywhere so we
        need to add these transitions to all states configured so far."""
        states = set()
        for state_name in self.state.known_states():
            if state_name not in ['new', 'finished']:
                states.add(state_name)
        for state in states:
            self._add_state(state, 'NOOP',  state)
            self._add_state(state, 'HELP',  state)
            self._add_state(state, 'QUIT',  'finished')

    def _build_state_machine(self):
        self.state = StateMachine(initial_state='new')
        self._add_state('new',     'GREET', 'greeted')
        self._add_state('greeted', 'HELO',  'initialized')

        self._add_state('greeted', 'EHLO',  'initialized', operations=('set_esmtp',))

        # ----
        self._add_state('initialized', 'MAIL FROM',  'sender_known')

        self._add_state('initialized', 'AUTH PLAIN', 'authenticated', condition='if_esmtp')
        self._add_state('authenticated', 'MAIL FROM',  'sender_known')
        # ----

        self._add_state('sender_known', 'RCPT TO',  'recipient_known')
        # multiple recipients
        self._add_state('recipient_known', 'RCPT TO',  'recipient_known')
        self._add_state('recipient_known', 'DATA',  'receiving_message')
        self._add_state('receiving_message', 'MSGDATA',  'initialized')
        self._add_help_noop_and_quit_transitions()
        self._add_rset_transitions()
        self.valid_commands = self.state.known_actions()

    # -------------------------------------------------------------------------

    def get_ehlo_lines(self):
        """Return the capabilities to be advertised after EHLO."""
        lines = []
        if self._authenticator != None:
            # TODO: Make the authentication pluggable but separate mechanism
            # from user look-up.
            lines.append('AUTH PLAIN')
        if self._policy != None:
            lines.extend(self._policy.ehlo_lines(self._message.peer))
        lines.append('HELP')
        return lines

    def _set_size_restrictions(self):
        """Set the maximum allowed message in the underlying layer so that big
        messages are not hold in memory before they are rejected."""
        self._command_parser.set_maximum_message_size(self._max_message_size())

    def _dispatch_commands(self, from_state, to_state, smtp_command):
        """This method dispatches a SMTP command to the appropriate handler
        method. It is called after a new command was received and a valid
        transition was found."""
        #print from_state, ' -> ', to_state, ':', smtp_command
        name_handler_method = 'smtp_%s' % smtp_command.lower().replace(' ', '_')
        try:
            handler_method = getattr(self, name_handler_method)
        except AttributeError:
            # base_msg = 'No handler for %s though transition is defined (no method %s)'
            # print base_msg % (smtp_command, name_handler_method)
            self.reply(451, 'Temporary Local Problem: Please come back later')
        else:
            # Don't catch InvalidDataError here - else the state would be moved
            # forward. Instead the handle_input will catch it and send out the
            # appropriate reply.
            handler_method()

    def _evaluate_decision(self, decision):
        return (decision in [True, None])

    def _is_multiline_reply(self, reply_message):
        return (not isinstance(reply_message, basestring))

    def _send_custom_response(self, reply):
        code, custom_response = reply
        if self._is_multiline_reply(custom_response):
            self.multiline_reply(code, custom_response)
        else:
            self.reply(code, custom_response)

    def _evaluate_policydecision_result(self, result):
        decision = self._evaluate_decision(result.is_command_acceptable())
        response_sent = result.use_custom_reply()
        if result.close_connection_before_response():
            self.close_connection()
            response_sent = True
        if result.use_custom_reply():
            self._send_custom_response(result.get_custom_reply())
        if result.close_connection_after_response():
            self.please_close_connection_after_response()
        return decision, response_sent

    def is_allowed(self, acl_name, *args):
        if self._policy is not None:
            decider = getattr(self._policy, acl_name)
            result = decider(*args)
            if result in [True, False, None]:
                return self._evaluate_decision(result), False
            elif hasattr(result, 'is_command_acceptable'):
                return self._evaluate_policydecision_result(result)
            elif len(result) == 2:
                decision = self._evaluate_decision(result[0])
                self._send_custom_response(result[1])
                return decision, True
            raise ValueError('Unknown policy response')
        return True, False

    # -------------------------------------------------------------------------

    def new_connection(self, remote_ip, remote_port):
        """This method is called when a new SMTP session is opened.
        [PUBLIC API]
        """
        self.state.set_state('new')
        self._message = Message(Peer(remote_ip, remote_port))

        decision, response_sent = self.is_allowed('accept_new_connection',
                                                  self._message.peer)
        if decision:
            if not response_sent:
                self.handle_input('greet')
            self._set_size_restrictions()
        else:
            if not response_sent:
                self.reply(554, 'SMTP service not available')
            self.close_connection()

    def handle_input(self, smtp_command, data=None):
        """Processes the given SMTP command with the (optional data).
        [PUBLIC API]
        """
        self._command_arguments = data
        self.please_close_connection_after_response(False)
        # SMTP commands must be treated as case-insensitive
        command = smtp_command.upper()
        try:
            try:
                self.state.execute(command)
            except StateMachineError:
                if command not in self.valid_commands:
                    self.reply(500, 'unrecognized command "%s"' % smtp_command)
                else:
                    msg = 'Command "%s" is not allowed here' % smtp_command
                    allowed_transitions = self.state.allowed_actions()
                    if len(allowed_transitions) > 0:
                        msg += ', expected on of %s' % allowed_transitions
                        self.reply(503, msg)
            except InvalidDataError:
                e = sys.exc_info()[1]
                self.reply(501, e.msg())
            except InvalidParametersError:
                # TODO: Get rid of InvalidParametersError, shouldn't be
                # necessary anymore
                e = sys.exc_info()[1]
                if not e.response_sent:
                    msg = 'Syntactically invalid %s argument(s)' % smtp_command
                    self.reply(501, msg)
            except PolicyDenial:
                e = sys.exc_info()[1]
                if not e.response_sent:
                    self.reply(e.code, e.reply_text)
        finally:
            if self.should_close_connection_after_response():
                self.close_connection()
            self._command_arguments = None

    def input_exceeds_limits(self):
        """Called when the client sent a message that exceeded the maximum
        size."""
        self.reply(552, 'message exceeds fixed maximum message size')

    def reply(self, code, text):
        """This method returns a message to the client (actually the session
        object is responsible of actually pushing the bits)."""
        self._command_parser.push(code, text)

    def multiline_reply(self, code, responses):
        """This method returns a message with multiple lines to the client
        (actually the session object is responsible of actually pushing the
        bits)."""
        self._command_parser.multiline_push(code, responses)

    def please_close_connection_after_response(self, value=None):
        if value is None:
            value = True
        self._close_connection_after_response = value

    def should_close_connection_after_response(self):
        return self._close_connection_after_response

    def close_connection(self):
        "Request a connection close from the SMTP session handling instance."
        if self._is_connected:
            self._is_connected = False
            self._command_parser.close_when_done()


    # -------------------------------------------------------------------------
    # Protocol handling functions (not public)

    def arguments(self):
        """Return the given parameters for the command as a string or an empty
        string"""
        return self._command_arguments or ''

    def smtp_greet(self):
        """This method handles not a real smtp command. It is called when a new
        connection was accepted by the server."""
        # Policy check was done when accepting the connection so we don't have
        # to do it here again.
        primary_hostname = self._command_parser.primary_hostname
        reply_text = '%s Hello %s' % (primary_hostname, self._message.peer.remote_ip)
        self.reply(220, reply_text)

    def validate(self, schema_class):
        context = dict(esmtp=self.uses_esmtp())
        return schema_class().process(self.arguments(), context=context)

    def smtp_quit(self):
        self.validate(SMTPCommandArgumentsSchema)
        primary_hostname = self._command_parser.primary_hostname
        reply_text = '%s closing connection' % primary_hostname
        self.reply(221, reply_text)
        self._command_parser.close_when_done()

    def smtp_noop(self):
        self.validate(SMTPCommandArgumentsSchema)
        self.reply(250, 'OK')

    def smtp_help(self):
        # deliberately no checking for additional parameters because RFC 821
        # says:
        # "The command may take an argument (e.g., any command name) and
        #  return more specific information as a response."
        states = self.get_all_allowed_smtp_commands()
        self.multiline_reply(214, ('Commands supported', ' '.join(states)))

    def _reply_to_helo(self, helo_string, response_sent):
        self._message.smtp_helo = helo_string
        if not response_sent:
            primary_hostname = self._command_parser.primary_hostname
            self.reply(250, primary_hostname)

    def _process_helo_or_ehlo(self, policy_methodname, reply_method):
        validated_data = self.validate(HeloSchema)
        helo_string = validated_data['helo']
        decision, response_sent = self.is_allowed(policy_methodname, helo_string, self._message)
        if decision:
            reply_method(helo_string, response_sent)
        elif not decision:
            raise PolicyDenial(response_sent)

    def smtp_helo(self):
        self._process_helo_or_ehlo('accept_helo', self._reply_to_helo)

    def _reply_to_ehlo(self, helo_string, response_sent):
        self._message.smtp_helo = helo_string
        if not response_sent:
            primary_hostname = self._command_parser.primary_hostname
            lines = [primary_hostname] + self.get_ehlo_lines()
            self.multiline_reply(250, lines)

    def smtp_ehlo(self):
        self._process_helo_or_ehlo('accept_ehlo', self._reply_to_ehlo)

    def _check_password(self, username, password):
        decision, response_sent = self.is_allowed('accept_auth_plain', username, password, self._message)
        if not decision:
            raise PolicyDenial(response_sent)
        assert response_sent == False
        credentials_correct = \
            self._authenticator.authenticate(username, password, self._message.peer)
        if credentials_correct:
            self._message.username = username
            self.reply(235, 'Authentication successful')
        else:
            self.reply(535, 'Bad username or password')

    def smtp_auth_plain(self):
        if self._authenticator is None:
            self.reply(535, 'AUTH not available')
            raise InvalidParametersError(response_sent=True)
        validated_data = self.validate(AuthPlainSchema)
        self._check_password(validated_data['username'], validated_data['password'])

    def _check_size_restriction(self, extensions):
        announced_size = extensions.get('size')
        if announced_size is None:
            return
        max_message_size = self._max_message_size()
        if max_message_size is None:
            return
        if announced_size > max_message_size:
            self.reply(552, 'message exceeds fixed maximum message size')
            raise InvalidParametersError('MAIL FROM', response_sent=True)

    def uses_esmtp(self):
        return self.state.is_set('esmtp')

    def smtp_mail_from(self):
        validated_data = self.validate(MailFromSchema)
        sender = validated_data['email']
        self._check_size_restriction(validated_data)
        decision, response_sent = self.is_allowed('accept_from', sender, self._message)
        if not decision:
            raise PolicyDenial(response_sent)
        self._message.smtp_from = sender
        if not response_sent:
            self.reply(250, 'OK')

    def smtp_rcpt_to(self):
        validated_data = self.validate(RcptToSchema)
        email_address = validated_data['email']
        decision, response_sent = self.is_allowed('accept_rcpt_to', email_address, self._message)
        if decision:
            self._message.smtp_to.append(email_address)
            if not response_sent:
                self.reply(250, 'OK')
        elif not decision:
            raise PolicyDenial(response_sent, 550, 'relay not permitted')

    def smtp_data(self):
        self.validate(SMTPCommandArgumentsSchema)
        decision, response_sent = self.is_allowed('accept_data', self._message)
        if decision and not response_sent:
            self._command_parser.switch_to_data_mode()
            self.reply(354, 'Enter message, ending with "." on a line by itself')
        elif not decision:
            raise PolicyDenial(response_sent)

    def _max_message_size(self):
        max_message_size = None
        if (self._policy is not None) and (self._message.peer is not None):
            max_message_size = self._policy.max_message_size(self._message.peer)
        return max_message_size

    def _check_size_restrictions(self, msg_data):
        max_message_size = self._max_message_size()
        if max_message_size is None:
            return
        msg_too_big = (len(msg_data) > int(max_message_size))
        if msg_too_big:
            msg = 'message exceeds fixed maximum message size'
            raise PolicyDenial(False, 552, msg)

    def _copy_basic_settings(self, msg):
        peer = self._message.peer
        new_message = Message(peer=Peer(peer.remote_ip, peer.remote_port),
                              smtp_helo=self._message.smtp_helo,
                              username=self._message.username)
        return new_message

    def smtp_msgdata(self):
        """This method handles not a real smtp command. It is called when the
        whole message was received (multi-line DATA command is completed)."""
        msg_data = self.arguments()
        self._command_parser.switch_to_command_mode()
        self._check_size_restrictions(msg_data)
        decision, response_sent = self.is_allowed('accept_msgdata', msg_data, self._message)
        if decision:
            self._message.msg_data = msg_data
            new_message = self._copy_basic_settings(self._message)
            self._deliverer.new_message_accepted(self._message)
            if not response_sent:
                self.reply(250, 'OK')
                # Now we must not loose the message anymore!
            self._message = new_message
        elif not decision:
            raise PolicyDenial(response_sent, 550, 'Message content is not acceptable')

    def smtp_rset(self):
        self.validate(SMTPCommandArgumentsSchema)
        self._message = Message(peer=self._message.peer,
                                smtp_helo=self._message.smtp_helo)
        self.reply(250, 'Reset OK')





#!/usr/bin/env python
# -*- coding: utf-8 -*-

import re

from setuptools import setup, find_packages


def requires_from_file(filename):
    requirements = []
    with open(filename, 'r') as requirements_fp:
        for line in requirements_fp.readlines():
            match = re.search('^\s*([a-zA-Z][^#]+?)(\s*#.+)?\n$', line)
            if match:
                requirements.append(match.group(1))
    return requirements

setup(
    name='pymta',
    version='0.6.0',
    description='library to build a custom SMTP server',
    long_description="""
pymta is a library to build a custom SMTP server in Python. This is useful if
you want to...

* test mail-sending code against a real SMTP server even in your unit tests.
* build a custom SMTP server with non-standard behavior without reimplementing
  the whole SMTP protocol.
* have a low-volume SMTP server which can be easily extended using Python""",
    packages = find_packages(),
    license='MIT',
    author='Felix Schwarz',
    author_email='felix.schwarz@oss.schwarz.eu',
    url='https://github.com/FelixSchwarz/pymta',
    install_requires=requires_from_file('requirements.txt'),
    tests_require=['nose'],
    test_suite='nose.collector',

    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Communications :: Email',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)



# -*- coding: UTF-8 -*-
# SPDX-License-Identifier: MIT
"""This module provides a unified view on certain symbols that are not present
for all versions of Python."""

try:
    import queue
except ImportError:
    # Python 2
    import Queue as queue
import sys
import base64

__all__ = ['queue']


if sys.version_info < (3, 0):
    basestring = basestring
    binary = bytes = str
    unicode = unicode
    range = xrange
    b64encode = lambda x: str(x).encode('base64')
    b64decode = lambda x: str(x).decode('base64')
else:
    basestring = str
    binary = bytes = b = bytes
    unicode = str
    range = range
    b64encode = lambda x: base64.b64encode(b(x)).decode('ascii')
    b64decode = lambda x: base64.b64decode(b(x)).decode('ascii')


def b(x, encoding='iso-8859-1'):
    if isinstance(x, unicode):
        return x.encode(encoding)
    return bytes(x)



# -*- coding: UTF-8 -*-
# SPDX-License-Identifier: MIT

from pymta.api import IMTAPolicy, PolicyDecision
from pymta.compat import b64encode

from tests.util import CommandParserTestCase, DummyAuthenticator


class BasicPolicyTest(CommandParserTestCase):
    "Tests that all commands can be controlled with policies."

    def test_connection_can_be_rejected(self):
        class FalsePolicy(IMTAPolicy):
            def accept_new_connection(self, peer):
                return False
        self.init(policy=FalsePolicy())
        self.assert_false(self.command_parser.open)
        self.assert_equals(1, len(self.command_parser.replies))
        code, reply_text = self.command_parser.replies[-1]
        self.assert_equals(554, code)
        self.assert_equals('SMTP service not available', reply_text)


    def test_helo_can_be_rejected(self):
        class FalsePolicy(IMTAPolicy):
            def accept_helo(self, helo_string, message):
                return (not helo_string.endswith('example.com'))
        self.init(policy=FalsePolicy())
        self.send('HELO', 'foo.example.com', expected_first_digit=5)
        self.send('HELO', 'bar.example.com', expected_first_digit=5)
        self.send('HELO', 'localhost')


    def test_ehlo_can_be_rejected(self):
        class FalsePolicy(IMTAPolicy):
            def accept_ehlo(self, ehlo_string, message):
                return (not ehlo_string.endswith('example.com'))
        self.init(policy=FalsePolicy())
        self.send('EHLO', 'foo.example.com', expected_first_digit=5)
        self.send('EHLO', 'localhost')


    def test_auth_plain_can_be_rejected(self):
        class FalsePolicy(IMTAPolicy):
            def accept_auth_plain(self, username, password, message):
                return False
        self.init(policy=FalsePolicy(), authenticator=DummyAuthenticator())
        self.send('EHLO', 'foo.example.com')
        base64_credentials = b64encode('\x00foo\x00foo')
        self.send('AUTH PLAIN', base64_credentials, expected_first_digit=5)


    def test_from_can_be_rejected(self):
        class FalsePolicy(IMTAPolicy):
            def accept_from(self, sender, message):
                return False
        self.init(policy=FalsePolicy())
        self.send('HELO', 'foo.example.com')
        self.send('MAIL FROM', 'foo@example.com', expected_first_digit=5)


    def test_rcptto_can_be_rejected(self):
        class FalsePolicy(IMTAPolicy):
            def accept_rcpt_to(self, new_recipient, message):
                return False
        self.init(policy=FalsePolicy())
        self.send('HELO', 'foo.example.com')
        self.send('MAIL FROM', 'foo@example.com')
        self.send('RCPT TO', 'to@example.com', expected_first_digit=5)


    def test_data_can_be_rejected(self):
        class FalsePolicy(IMTAPolicy):
            def accept_data(self, message):
                return False
        self.init(policy=FalsePolicy())
        self.send('HELO', 'foo.example.com')
        self.send('MAIL FROM', 'foo@example.com')
        self.send('RCPT TO', 'to@example.com')
        self.send('DATA', expected_first_digit=5)


    def test_messages_can_be_rejected(self):
        class FalsePolicy(IMTAPolicy):
            def accept_msgdata(self, msg_data, message):
                return False
        self.init(policy=FalsePolicy())
        self.send('HELO', 'foo.example.com')
        self.send('MAIL FROM', 'foo@example.com')
        self.send('RCPT TO', 'to@example.com')
        self.send('DATA', expected_first_digit=3)
        rfc822_msg = 'Subject: Test\n\nJust testing...\n'
        self.send('MSGDATA', rfc822_msg, expected_first_digit=5)


    def test_size_limit_messages_can_be_rejected(self):
        class MaxSizePolicy(IMTAPolicy):
            def max_message_size(self, peer):
                return 100
        self.init(policy=MaxSizePolicy())
        self.send('HELO', 'foo.example.com')
        self.send('MAIL FROM', 'foo@example.com')
        self.send('RCPT TO', 'to@example.com')
        self.send('DATA', expected_first_digit=3)
        big_data_chunk = ('x'*70 + '\n') * 1500
        rfc822_msg = 'Subject: Test\n\nJust testing...\n' + big_data_chunk
        (code, reply_text) = self.send('MSGDATA', rfc822_msg,
                                       expected_first_digit=5)
        self.assert_equals(552, code)
        self.assert_equals('message exceeds fixed maximum message size', reply_text)

    def test_server_deals_gracefully_with_double_close_because_of_faulty_policy(self):
        class DoubleCloseConnectionPolicy(IMTAPolicy):
            def accept_helo(self, helo_string, message):
                decision = PolicyDecision(True)
                decision._close_connection_before_response = True
                decision._close_connection_after_response = True
                return decision
        self.init(policy=DoubleCloseConnectionPolicy())

        number_replies_before = len(self.command_parser.replies)
        self.session.handle_input('HELO', 'foo.example.com')
        self.assert_equals(number_replies_before, len(self.command_parser.replies))
        self.assert_false(self.command_parser.open)




# -*- coding: UTF-8 -*-
# SPDX-License-Identifier: MIT

from pymta.compat import b64encode

from tests.util import CommandParserTestCase, DummyAuthenticator


class SMTPParameterValidationTest(CommandParserTestCase):

    def last_server_message(self):
        last_code, last_message = self.command_parser.replies[-1]
        return last_message

    def send_invalid(self, command, data=None):
        return super(SMTPParameterValidationTest, self).send(command, data=data, expected_first_digit=5)

    def send_valid(self, command, data=None):
        return super(SMTPParameterValidationTest, self).send(command, data=data, expected_first_digit=2)

    # -------------------------------------------------------------------------
    # helo/ehlo

    def test_helo_accepts_exactly_one_parameter(self):
        self.send_invalid('helo')
        self.send_invalid('helo', 'foo bar')
        self.send_invalid('helo', '')

    def test_ehlo_accepts_exactly_one_parameter(self):
        self.send_invalid('ehlo')
        self.send_invalid('ehlo', 'foo bar')

    # -------------------------------------------------------------------------
    # commands without parameters

    def helo(self):
        self.send_valid('helo', 'fnord')

    def test_noop_does_not_accept_any_parameters(self):
        self.helo()

        self.send_invalid('noop', 'foo')

    def test_rset_does_not_accept_any_parameters(self):
        self.helo()

        self.send_invalid('rset', 'foo')

    def test_quit_does_not_accept_any_parameters(self):
        self.helo()

        self.send_invalid('quit', 'invalid')

    def test_data_does_not_accept_any_parameters(self):
        self.helo_and_mail_from()
        self.send_valid('rcpt to', 'foo@example.com')

        self.send_invalid('data', 'invalid')

    # -------------------------------------------------------------------------
    # MAIL FROM

    def test_mail_from_requires_an_email_address(self):
        self.helo()

        self.send_invalid('mail from')
        self.send_invalid('mail from', 'foo@@bar')

    def test_mail_from_must_not_have_extensions_for_plain_smtp(self):
        self.helo()

        self.send_invalid('mail from', '<foo@example.com> SIZE=100')
        self.assert_equals('No SMTP extensions allowed for plain SMTP.', self.last_server_message())

    def ehlo(self):
        self.send_valid('ehlo', 'fnord')

    def test_mail_from_validates_size_extension(self):
        self.ehlo()

        self.send_invalid('mail from', '<foo@example.com> SIZE=fnord')

    def test_mail_from_rejects_unknown_extension(self):
        self.send_valid('ehlo', 'fnord')

        self.send_invalid('mail from', '<foo@example.com> FNORD=INVALID')
        self.assert_equals("Invalid extension: 'FNORD=INVALID'", self.last_server_message())

    def helo_and_mail_from(self):
        self.helo()
        self.send_valid('mail from', 'foo@example.com')

    # -------------------------------------------------------------------------
    # RCPT TO

    def test_rcpt_to_requires_an_email_address(self):
        self.helo_and_mail_from()

        self.send_invalid('rcpt to')
        self.send_invalid('rcpt to foo@@bar.com')
        self.send_invalid('rcpt to foo@bar.com invalid')

    def test_rcpt_to_accepts_a_valid_email_address(self):
        self.helo_and_mail_from()
        self.send_valid('rcpt to', 'foo@example.com')
        self.send_valid('rcpt to', '<foo@example.com>')

    # -------------------------------------------------------------------------
    # AUTH PLAIN

    def inject_authenticator(self):
        self.session._authenticator = DummyAuthenticator()

    def base64(self, value):
        return b64encode(value).strip()

    def test_auth_plain_accepts_correct_authentication(self):
        self.inject_authenticator()
        self.ehlo()

        self.send_valid('AUTH PLAIN', b64encode('\x00foo\x00foo'))

    def test_auth_plain_requires_exactly_one_parameter(self):
        self.inject_authenticator()
        self.ehlo()

        self.send_invalid('AUTH PLAIN')
        base64_credentials = self.base64('\x00foo\x00foo')
        self.send_invalid('AUTH PLAIN', base64_credentials + ' ' + base64_credentials)

    def test_auth_plain_detects_bad_base64_credentials(self):
        self.inject_authenticator()
        self.ehlo()

        self.send_invalid('AUTH PLAIN')
        self.send_invalid('AUTH PLAIN', 'invalid_base64')

    def test_auth_plain_reject_bad_credentials(self):
        self.inject_authenticator()
        self.ehlo()

        self.send_invalid('AUTH PLAIN', self.base64('\x00foo\x00bar'))





# -*- coding: UTF-8 -*-
# SPDX-License-Identifier: MIT

import re


__all__ = ['StateMachine', 'StateMachineDefinitionError', 'StateMachineError']


class StateMachineError(Exception):
    pass


class StateMachineDefinitionError(StateMachineError):
    pass


class StateMachine(object):
    def __init__(self, initial_state=None):
        self._state = initial_state
        self._transitions = {}
        self._flags = {}

    # --- states ---------------------------

    def state(self):
        if self.is_impossible_state():
            return None
        return self._state

    def is_impossible_state(self):
        return (self._state not in self.known_states())

    def set_state(self, state):
        if state not in self.known_states():
            raise StateMachineError
        self._state = state

    # --- transitions ----------------------

    def add(self, from_state, to_state, action_name, handler=None, operations=(), condition=None):
        #print 'from_state, to_state, action_name', repr((from_state, to_state, action_name))
        self._transitions.setdefault(from_state, {})
        if action_name in self._transitions[from_state]:
            old_to_state = self._transitions[from_state][action_name][0]
            msg = 'Duplicate action "%s" for state "%s" (-> "%s" already known, can not add transition to "%s")' % (action_name, from_state, old_to_state, to_state)
            raise StateMachineDefinitionError(msg)
        self._transitions[from_state][action_name] = (to_state, handler, operations, condition)

    def execute(self, action_name):
        if action_name not in self.allowed_actions():
            msg = 'Invalid action "%s", expected one of %s' % (action_name, self.allowed_actions())
            raise StateMachineError(msg)

        current_state = self.state()
        current_transitions = self._transitions.get(current_state, {})
        final_state, handler, operations, condition = current_transitions[action_name]
        if handler is not None:
            handler(current_state, final_state, action_name)
        for operation in operations:
            self._execute_operation(operation)
        self._state = final_state

    # --- flags ----------------------------

    def is_set(self, flag):
        return self._flags.get(flag, False)

    def _execute_operation(self, operation):
        match = re.search('^set_(\w+)$', operation)
        assert match is not None
        flag_name = match.group(1)
        self._flags[flag_name] = True

    def _is_condition_satisfied(self, condition):
        if condition is None:
            return True
        match = re.search('^if_(not_)?(\w+?)$', condition)
        assert match is not None
        flag = match.group(2)
        if match.group(1):
            return not self.is_set(flag)
        return self.is_set(flag)

    # --- introspection --------------------

    def known_actions(self):
        actions = set()
        for action_name in self._transitions.values():
            actions = actions.union(action_name)
        return actions

    def allowed_actions(self):
        current_transitions = self._transitions.get(self.state(), {})
        _allowed_actions = set()
        for action_name, (to_state, handler, operations, condition) in current_transitions.items():
            if not self._is_condition_satisfied(condition):
                continue
            _allowed_actions.add(action_name)
        return _allowed_actions

    def known_non_final_states(self):
        return set(self._transitions)

    def known_states(self):
        states = self.known_non_final_states()
        for from_state, action_names in self._transitions.items():
            for action_name in action_names:
                (to_state, handler, operations, condition) = self._transitions[from_state][action_name]
                states.update((to_state,))
        return states





# -*- coding: UTF-8 -*-
# SPDX-License-Identifier: MIT

import smtplib
import socket
import time

from pythonic_testcase import *

from pymta.api import IMTAPolicy
from pymta.test_util import DebuggingMTA, SMTPTestCase
from pymta.compat import b

from tests.util import DummyAuthenticator


rfc822_msg = 'Subject: Test\n\nJust testing...'



class BasicSMTPTest(SMTPTestCase):
    """This test uses the SMTP protocol to check the whole server."""

    def build_mta(self, hostname, listen_port, deliverer, policy_class=None):
        return DebuggingMTA(hostname, listen_port, deliverer,
                            authenticator_class=DummyAuthenticator,
                            policy_class=policy_class)

    def init_mta(self, policy_class=IMTAPolicy):
        super(BasicSMTPTest, self).init_mta(policy_class)
        self.connection = smtplib.SMTP()
        self.connection.set_debuglevel(0)
        self.connection.connect(self.hostname, self.listen_port)

    def stop_mta(self):
        if getattr(self, 'connection', None) is not None:
            try:
                self.connection.quit()
            except smtplib.SMTPServerDisconnected:
                pass
        super(BasicSMTPTest, self).stop_mta()

    def test_helo(self):
        code, replytext = self.connection.helo('foo')
        self.assert_equals(250, code)

    def _check_received_mail(self, expected_recipients, expected_helo=None):
        queue = self.get_received_messages()
        self.assert_equals(1, queue.qsize())
        msg = queue.get()
        if expected_helo is not None:
            self.assert_equals('foo', msg.smtp_helo)
        self.assert_equals('from@example.com', msg.smtp_from)
        self.assert_equals(expected_recipients, msg.smtp_to)
        self.assert_equals(rfc822_msg, msg.msg_data)
        return msg

    def test_send_simple_email(self):
        code, replytext = self.connection.helo('foo')
        self.assert_equals(250, code)
        code, replytext = self.connection.mail('from@example.com')
        self.assert_equals(250, code)
        code, replytext = self.connection.rcpt('to@example.com')
        self.assert_equals(250, code)
        code, replytext = self.connection.data(rfc822_msg)
        self.assert_equals(250, code)
        self.connection.quit()
        self._check_received_mail(['to@example.com'], expected_helo='foo')

    def test_send_email_via_smtplib(self):
        """Check that we can send a simple email via smtplib.sendmail without
        using the low-level api."""
        recipient = 'to@example.com'
        self.connection.sendmail('from@example.com', recipient, rfc822_msg)
        self.connection.quit()
        self._check_received_mail([recipient])

    def test_multiple_recipients(self):
        """Check that we can send an email to multiple recipients at once."""
        recipients = ['foo@example.com', 'bar@example.com']
        self.connection.sendmail('from@example.com', recipients, rfc822_msg)
        self.connection.quit()
        self._check_received_mail(recipients)

    def test_send_email_with_authentication(self):
        """Check that we can send an email with prior authentication."""
        recipient = 'to@example.com'
        self.connection.login('admin', 'admin')
        self.connection.sendmail('from@example.com', recipient, rfc822_msg)
        self.connection.quit()
        msg = self._check_received_mail([recipient])
        self.assert_equals('admin', msg.username)

    def test_send_multiple_emails_in_one_connection(self):
        """Check that we can send multiple emails in the same connection (and
        the second email needs to have the same peer information/helo string
        though this information is only sent once)."""
        self.connection.login('admin', 'admin')
        self.connection.sendmail('x@example.com', 'foo@example.com', rfc822_msg)
        self.connection.sendmail('x@example.com', 'bar@example.com', rfc822_msg)

        queue = self.get_received_messages()
        self.assert_equals(2, queue.qsize())
        first_msg = queue.get()
        self.assert_not_none(first_msg.smtp_helo)
        self.assert_equals('x@example.com', first_msg.smtp_from)

        second_msg = queue.get()
        self.assert_equals(first_msg.smtp_helo, second_msg.smtp_helo)
        self.assert_equals(first_msg.username, second_msg.username)

    def test_transparency_support_enabled(self):
        """Check that there is transparency support for lines starting with a
        dot in the message body (RFC 821, section 4.5.2)."""
        msg = rfc822_msg + '\n.Bar\nFoo'
        self.connection.sendmail('from@example.com', 'foo@example.com', msg)
        self.connection.quit()

        queue = self.get_received_messages()
        self.assert_equals(1, queue.qsize())
        received_msg = queue.get()
        self.assert_equals(msg, received_msg.msg_data)

    def test_big_messages_are_rejected(self):
        """Check that messages which exceed the configured maximum message size
        are rejected. This tests all the code setting the maximum allowed input
        size in the transport layer."""
        class RestrictedSizePolicy(IMTAPolicy):
            def max_message_size(self, peer):
                return 100

        self.stop_mta()
        self.init_mta(RestrictedSizePolicy)

        big_data_chunk = ('x'*70 + '\n') * 1500
        msg = rfc822_msg + big_data_chunk

        # Depending on when the protocol used (ESMTP with SIZE extension or
        # plain old SMTP) and when the server checks the message size (after
        # MAIL FROM or only when the message was transmitted completely),
        # the error message may differ.
        # Unfortunately Python's smtplib raises SMTPSenderRefused even if the
        # message was rejected due to size restrictions after issuing MAIL FROM
        # with size verb
        with assert_raises((smtplib.SMTPDataError, smtplib.SMTPSenderRefused)) as exc_state:
            self.connection.sendmail('from@example.com', 'foo@example.com', msg)
        e = exc_state.caught_exception
        self.assert_equals(552, e.smtp_code)
        self.assert_equals(b('message exceeds fixed maximum message size'), e.smtp_error)

    def service_is_available(self):
        # On a normal system we should be able to reconnect after a dropped
        # connection within two seconds under all circumstances.
        old_default = socket.getdefaulttimeout()
        socket.setdefaulttimeout(2)
        try:
            try:
                self.connection.connect(self.hostname, self.listen_port)
                return True
            finally:
                socket.setdefaulttimeout(old_default)
        except socket.timeout:
            return False

    def test_workerprocess_detects_closed_connections_when_reading(self):
        """Check that the WorkerProcess gracefully handles connections which are
        closed without QUIT. This can happen due to network problems or
        unfriendly clients."""
        self.connection.helo('foo')
        self.connection.close()

        # In 0.3 the WorkerProcess would hang and start to eat up the whole CPU
        # so we need to set a sensible timeout so that this test will fail with
        # an appropriate exception.
        self.assert_true(self.service_is_available())

    def test_workerprocess_detects_closed_connections_when_writing(self):
        """Check that the WorkerProcess gracefully handles connections which are
        closed without QUIT - remaining output is suppressed. This can happen
        due to network problems or unfriendly clients."""
        # Basically the problem also occurs without waiting - however sleeping a
        # bit increases the likelyhood to trigger to problem. We need to make
        # sure that the server writes to a already closed TCP connection.
        class WaitingPolicy(IMTAPolicy):
            def accept_helo(self, helo_string, message):
                is_first_time = getattr(self, 'is_first_time', False)
                if is_first_time:
                    time.sleep(1)
                # Also the socket will buffer some data - make sure the system
                # actually writes data to the socket so we get an exception.
                return (False, (552, ('Go away',)*10))
        self.init_mta(policy_class=WaitingPolicy)

        # don't wait for an answer as .helo() does
        self.connection.putcmd('helo', 'foo')
        self.connection.close()

        time.sleep(0.5)
        self.assert_true(self.service_is_available())
